"""Generation of metadata for the OME-NGFF (zarr) slides.

Based on version 0.4 of the specification:
https://ngff.openmicroscopy.org/0.4/
"""
import sys
from dataclasses import dataclass, field
from numbers import Number

# if python < 3.8 use typing extensions
if sys.version_info < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal

from typing import List, Optional, Union

import wsic

SpaceUnits = Literal[
    "angstrom",
    "attometer",
    "centimeter",
    "decimeter",
    "exameter",
    "femtometer",
    "foot",
    "gigameter",
    "hectometer",
    "inch",
    "kilometer",
    "megameter",
    "meter",
    "micrometer",
    "mile",
    "millimeter",
    "nanometer",
    "parsec",
    "petameter",
    "picometer",
    "terameter",
    "yard",
    "yoctometer",
    "yottameter",
    "zeptometer",
    "zettameter",
]
TimeUnits = Literal[
    "attosecond",
    "centisecond",
    "day",
    "decisecond",
    "exasecond",
    "femtosecond",
    "gigasecond",
    "hectosecond",
    "hour",
    "kilosecond",
    "megasecond",
    "microsecond",
    "millisecond",
    "minute",
    "nanosecond",
    "petasecond",
    "picosecond",
    "second",
    "terasecond",
    "yoctosecond",
    "yottasecond",
    "zeptosecond",
    "zettasecond",
]
TCZYX = Literal["t", "c", "z", "y", "x"]


@dataclass
class Creator:
    """Record the creator (wsic) information.

    Attributes:
        name (str):
            The name of the creator.
        version (str):
            The version of the creator.
    """

    name: str = "wsic"
    version: str = wsic.__version__


@dataclass
class CoordinateTransform:
    """Transformation from the zarr to slide coordinate system.

    Attributes:
        type (str):
            The type of coordinate transform. E.g. "scale".
        scale (List[float]):
            The scale factors. Must be one for each axis.

    """

    type: str = "scale"  # noqa: A003
    scale: List[float] = field(default_factory=[1.0, 0.5, 0.5])


@dataclass
class Dataset:
    """Description of a single resolution.

    Attributes:
        path (str):
            Path to the dataset. This will usually be a string of an
            integer e.g. "0".
        coordinateTransformations (List[CoordinateTransform]):
            Transformations from the zarr to slide coordinate system.
    """

    path: str = "0"
    coordinateTransformations: List[CoordinateTransform] = field(  # noqa: N815
        default_factory=lambda: [CoordinateTransform()]
    )


@dataclass
class Axis:
    """Description of an axis including type and units.

    Attributes:
        name (str):
            The name of the axis. Must be one of: "t", "c", "z", "y",
            "x".
        type (str):
            The type of the axis. Msut be one of: "time", "space",
            "channel".
        unit (str):
            The units of the axis.
    """

    name: TCZYX
    type: Literal["time", "space", "channel"]  # noqa: A003
    unit: Optional[Union[SpaceUnits, TimeUnits]] = None


@dataclass
class Multiscales:
    """Description of multiple resolutions present.

    Attributes:
        axes (List[Axis]):
            The axes of the multiscales.
        datasets (List[Dataset]):
            The datasets of the multiscales.
        version (str):
            The version of the specification.
    """

    axes: List[Axis] = field(
        default_factory=lambda: [
            Axis("y", "space", "micrometer"),
            Axis("x", "space", "micrometer"),
            Axis("c", "channel", None),
        ]
    )
    datasets: List[Dataset] = field(default_factory=lambda: [Dataset()])
    version: str = "0.4"


@dataclass
class Window:
    """The range of values within a channel.

    Attributes:
        end (int):
            The end of the window.
        max (int):
            The maximum value in the window.
        min (int):
            The minimum value in the window.
        start (int):
            The start of the window.
    """

    end: Number = 255
    max: Number = 255  # noqa: A003
    min: Number = 0  # noqa: A003
    start: Number = 0


@dataclass
class Channel:
    """Description of a single channel.

    Attributes:
        active (bool):
            Whether the channel is active by default.
        color (str):
            The color of the channel in hexadecimal format. E.g.
            "FF0000" for red.
        family (str):
            The family of the channel. E.g. "linear".
        inverted (bool):
            Whether the channel is inverted.
        window (Window):
            The min and max values represented in the channel.
    """

    active: bool = True
    coefficient: float = 1.0
    color: str = "FF0000"  # Hexadecimal color code
    family: str = "linear"
    inverted: bool = False
    label: str = "Red"
    window: Window = field(default_factory=Window)


@dataclass
class RDefs:
    """Defaults for axes and colour model.

    Attributes:
        defaultT (int):
            Default timepoint.
        defaultZ (int):
            Default z-plane.
        model (str):
            Colour model: "color" or "greyscale".
    """

    defaultT: int = 0  # noqa: N815
    defaultZ: int = 0  # noqa: N815
    model: Literal["color", "greyscale"] = "color"


@dataclass
class Omero:
    """Display information e.g. colour channel information.

    Attributes:
        name (str):
            The display name.
        id (int):
            The omero ID.
        channels (List[Channel]):
            The colour channels.
        rdefs (RDefs):
            The default values for axes and colour model.
        version (str):
            The version of the specification.
    """

    name: Optional[str] = None
    id: int = 1  # noqa: A003
    channels: list = field(
        default_factory=lambda: [
            Channel(label="Red", color="FF0000"),
            Channel(label="Green", color="00FF00"),
            Channel(label="Blue", color="0000FF"),
        ]
    )
    rdefs: RDefs = field(default_factory=RDefs)
    version: str = "0.4"


@dataclass
class Zattrs:
    """Root metadata.

    Attributes:
        _creator (Creator):
            Information about the creator.
        multiscales (Multiscales):
            Information about the multiscales.
        _ARRAY_DIMENSIONS (List[TCZYX]):
            The dimensions of the array, for xarray compatibility.
        omero (Omero):
            Information about the display of image data.
    """

    _creator: Creator = field(default_factory=Creator)
    multiscales: List[Multiscales] = field(default_factory=lambda: [Multiscales()])
    _ARRAY_DIMENSIONS: List[TCZYX] = field(default_factory=lambda: ["y", "x", "c"])
    omero: Omero = field(default_factory=Omero)



import multiprocessing.queues
from typing import Any


class Queue(multiprocessing.queues.Queue):
    """A multiprocessing.Queue subclass with a shared counter.

    The `multiprocessing.Queue.qsize` is unreliable and may raise
    `NotImplementedError` on Unix platforms like macOS where
    `sem_getvalue()` is not implemented. Here we use a shared counter
    instead.
    """

    def __init__(self, *args, **kwargs) -> None:
        """Initialise the queue."""
        super().__init__(*args, ctx=multiprocessing.get_context(), **kwargs)
        self.counter = multiprocessing.Value("i", 0)

    def put(self, *args, **kwargs) -> None:
        """Put an item into the queue."""
        with self.counter.get_lock():
            self.counter.value += 1
        return super().put(*args, **kwargs)

    def get(self, *args, **kwargs) -> Any:
        """Remove and return an item from the queue."""
        value = super().get(*args, **kwargs)
        with self.counter.get_lock():
            self.counter.value -= 1
        return value  # noqa: R504

    def qsize(self) -> int:
        """Return the number of items in the queue."""
        return self.counter.value

    def empty(self) -> bool:
        """Return True if the queue is empty."""
        return self.qsize() == 0

    def clear(self) -> None:
        """Clear the queue."""
        while not self.empty():
            self.get()

    def __getstate__(self) -> Any:
        """Return the state of the queue."""
        return (self.counter, super().__getstate__())

    def __setstate__(self, state: Any) -> None:
        """Restore the state of the queue."""
        self.counter, state = state
        super().__setstate__(state)
        self._after_fork()



"""Detect file type by searching for signatures (magic numbers)."""
import mmap
import re
from dataclasses import dataclass
from pathlib import Path
from typing import IO, List, Optional, Sequence, Tuple, Union

from wsic.types import Magic, PathLike


@dataclass
class Spell:
    """A magic number (or regex) and optional offset to search for in data.

    When the spell is 'performed' it will return True if the data
    contains the magic. Magic can be a bytes object (AKA a magic number)
    or a compiled regular expression.

    Args:
        magic (bytes or re.Pattern):
            The magic to search for.
        offset (int, optional):
            The offset to start searching from. None means that the
            magic may be anywhere.

    """

    magic: Magic
    offset: Optional[int] = 0

    def perform(self, data: Union[bytes, mmap.mmap]) -> bool:
        """Check if data contains magic.

        Args:
            data (bytes or mmap.mmap):
                The data to search.

        Returns:
            bool:
                True if the spell was successful, i.e. if the data
                contains magic.
        """
        if isinstance(self.magic, bytes) and self.offset is not None:
            return self.magic == data[self.offset : self.offset + len(self.magic)]
        if isinstance(self.magic, re.Pattern):
            return self.magic.search(data, pos=self.offset or 0) is not None
        return self.magic in data


@dataclass
class Incantation:
    r"""Perform a sequence of spells.

    Spells are performed in order and nesting of spells is supported.

    Nesting defines alternating conjunction and disjunction of spells.
    For example, ALL top level spell or sequence of spells must succeed,
    ANY one of the second level spell must succeed, etc.

    Examples:

    >>> # Both of the following spell must succeed to return True:
    >>> Incantation(
    ...     spells=[
    ...         Spell(b'\x0A'),
    ...         Spell(b'\x0B'),
    ...     ],
    ... )

    >>> # Either of the following spell must succeed to return True:
    >>> Incantation(
    ...     spells=[[
    ...         Spell(b'\x0A'),
    ...         Spell(b'\x0B'),
    ...     ]],
    ... )


    Args:
        spells (Sequence[Union[Sequence[Spell], Spell]]):
            The sequence of Spells to perform.
        data (Union[bytes, mmap.mmap]):
            The data to perform the incantation on.

    """

    spells: Sequence[Union[Sequence[Spell], Spell]]

    @staticmethod
    def disjunction(
        spells: Union[Spell, Sequence[Spell]], data: Union[bytes, mmap.mmap]
    ) -> bool:
        """Check that ANY spell succeeds."""
        if isinstance(spells, Spell):
            return spells.perform(data)
        return any(Incantation.conjunction(spell, data) for spell in spells)

    @staticmethod
    def conjunction(
        spells: Union[Spell, Sequence[Spell]], data: Union[bytes, mmap.mmap]
    ) -> bool:
        """Check that ALL spells succeed."""
        if isinstance(spells, Spell):
            return spells.perform(data)
        return all(Incantation.disjunction(spell, data) for spell in spells)

    def perform(self, data: Union[bytes, mmap.mmap]) -> bool:
        """Check that the incantation succeeds."""
        return Incantation.conjunction(self.spells, data)


# A mapping of file types to incantations.
FILE_INCANTATIONS = {
    ("tiff",): Incantation(
        spells=[
            (
                # Little-endian TIFF
                Spell(b"II\x2a\x00"),
                # Little-endian Big-TIFF
                Spell(b"II\x2b\x00"),
                # Big-endian TIFF
                Spell(b"MM\x00\x2a"),
                # Big-endian Big-TIFF
                Spell(b"MM\x00\x2b"),
            ),
        ],
    ),
    ("tiff", "svs"): Incantation(
        spells=[
            # "Aperio" appears in the description tag
            Spell(b"Aperio", None),
        ],
    ),
    ("jpeg",): Incantation(
        spells=[
            # JPEG start of image (SOI) marker
            Spell(b"\xff\xd8"),
        ],
    ),
    ("png",): Incantation(
        spells=[
            # PNG signature
            Spell(b"\x89PNG\r\n\x1a\n"),
        ],
    ),
    ("jp2",): Incantation(
        spells=[
            # JPEG 2000 signature
            Spell(b"\x00\x00\x00\x0cjP  \r\n\x87\n"),
        ],
    ),
    ("jp2", "omnyx"): Incantation(
        spells=[
            # "Omnyx" appears in the description tag of the XML box
            Spell(re.compile(b"<description>\\s*Omnyx", re.I), None),
        ],
    ),
    ("webp",): Incantation(
        spells=[
            # WEBP signature where .... is the file size
            Spell(re.compile(b"RIFF....WEBP")),
        ],
    ),
    ("dcm",): Incantation(
        spells=[
            # DICOM signature
            Spell(b"DICM", 128),
        ],
    ),
}


def _perform_dir_incantations(path: Path):
    """Detect matching types for a directory.

    E.g. a zarr group/array or a DICOM directory.
    """
    # Zarr
    zgroup = path / ".zgroup"
    if zgroup.exists():
        return [("zarr",), ("zarr", "group")]
    zarray = path / ".zarray"
    if zarray.exists():
        return [("zarr",), ("zarr", "array")]
    # DICOM
    for sub_path in path.iterdir():
        if sub_path.is_dir():
            continue
        with sub_path.open("rb") as file_handle:
            header = file_handle.read(128 + 4)
        if Spell(b"DICM", 128).perform(header):
            return [("dicom",)]
    return []


def header_bytes(
    header_length: Optional[int], file_handle: IO
) -> Union[bytes, mmap.mmap]:
    """Read in bytes from the start as a header for checking.

    If 0 or None is given for header_length, the entire file is memory
    mapped.

    Args:
        header_length (int):
            The number of bytes to read from the start of the file.
            If 0 or None, the entire file is memory mapped and checked.
        file_handle (file):
            The file handle to read from.

    Returns:
        The header bytes or memory mapped bytes.
    """
    if header_length:
        return file_handle.read(header_length)
    return mmap.mmap(file_handle.fileno(), 0, access=mmap.ACCESS_READ)


def summon_file_types(
    path: PathLike,
    header_length: Optional[int] = 1024,
) -> List[Tuple[str, ...]]:
    """Perform a series of incantations to determine the file types.

    A list of types is returned becuse the file may contain multiple
    magic numbers or patterns. E.g. a file may be

    Args:
        path (PathLike):
            The path to the file.
        header_length (int, optional):
            The number of bytes at the start of the file to read for
            checking. If None, the entire file is memory mapped and
            checked.

    Returns:
        A list of file types for which this file has the correct magic.
    """
    path = Path(path)
    file_types = []

    # Handle directory cases
    if path.is_dir():
        return _perform_dir_incantations(path)

    # Handle file cases
    with path.open("rb") as file_handle:
        header = header_bytes(header_length, file_handle)

        # Iterate over FILE_INCANTATIONS sorted by the type tuple length
        # to ensure that subtypes are checked after the parent type.
        # This avoids the incantaction for a subtype having to perform
        # the incantation for the parent type again.
        ordered_incantations = sorted(
            FILE_INCANTATIONS.items(), key=lambda x: len(x[0])
        )
        for file_type, incantation in ordered_incantations:
            parent_type = file_type[:-1]
            parent_type_matched = (parent_type in file_types) or (parent_type == ())
            if incantation.perform(header) and parent_type_matched:
                file_types.append(file_type)
    return file_types


def pentagram() -> None:
    """Print a pentagram."""
    print(
        """
            @@@@@@@@@@@@
        @@@@            @@@@
      @@ ##              ## @@
    @@   # ##          ## #   @@
  @@      #  ##      ##  #      @@
 @        #    ##  ##    #        @
 @         #     ##     #         @
@          #  ###  ###  #          @
@          ###        ###          @
@       ### #          # ###       @
@    ###     #        #     ###    @
@  ##        #        #        ##  @
@##################################@
 @            #      #            @
 @             #    #             @
  @@           #    #           @@
    @@          #  #          @@
      @@        #  #        @@
        @@@@     ##     @@@@
            @@@@@@@@@@@@
    """
    )



import multiprocessing
import os
import time
import warnings
from abc import ABC
from concurrent.futures import ThreadPoolExecutor
from contextlib import suppress
from math import ceil, floor
from pathlib import Path
from typing import Iterable, Iterator, Optional, Tuple, Union

import numpy as np
import zarr

from wsic.enums import Codec, ColorSpace
from wsic.magic import summon_file_types
from wsic.multiprocessing import Queue
from wsic.types import PathLike
from wsic.utils import (
    block_downsample_shape,
    mean_pool,
    mosaic_shape,
    ppu2mpp,
    resize_array,
    scale_to_fit,
    tile_slices,
    warn_unused,
    wrap_index,
)


class Reader(ABC):
    """Base class for readers."""

    def __init__(self, path: PathLike):
        """Initialize reader.

        Args:
            path (PathLike):
                Path to file.
        """
        self.path = Path(path)

    def __getitem__(self, index: Tuple[Union[int, slice], ...]) -> np.ndarray:
        """Get pixel data at index."""
        raise NotImplementedError

    @classmethod
    def from_file(cls, path: Path) -> "Reader":
        """Return reader for file.

        Args:
            path (Path): Path to file.

        Returns:
            Reader: Reader for file.
        """
        path = Path(path)
        file_types = summon_file_types(path)
        if ("jp2",) in file_types:
            return JP2Reader(path)
        with suppress(ImportError):
            import openslide

            with suppress(openslide.OpenSlideUnsupportedFormatError):
                return OpenSlideReader(path)
        if ("tiff",) in file_types:
            return TIFFReader(path)
        if ("dicom",) in file_types or ("dcm",) in file_types:
            return DICOMWSIReader(path)
        if ("zarr",) in file_types:
            from wsic.writers import ZarrReaderWriter

            return ZarrReaderWriter(path)
        raise ValueError(f"Unsupported file type: {path}")

    def thumbnail(self, shape: Tuple[int, ...], approx_ok: bool = False) -> np.ndarray:
        """Generate a thumbnail image of (or near) the requested shape.

        Args:
            shape (Tuple[int, ...]):
                Shape of the thumbnail.
            approx_ok (bool):
                If True, return a thumbnail that is approximately the
                requested shape. It will be equal to or larger than the
                requested shape (the next largest shape possible via
                an integer block downsampling).

        Returns:
            np.ndarray: Thumbnail.
        """
        # NOTE: Assuming first two are Y and X
        yx_shape = self.shape[:2]
        yx_tile_shape = (
            self.tile_shape[:2] if self.tile_shape else np.minimum(yx_shape, (256, 256))
        )
        self_mosaic_shape = (
            self.mosaic_shape
            if self.mosaic_shape
            else mosaic_shape(yx_shape, yx_tile_shape)
        )
        downsample_shape = yx_shape
        downsample_tile_shape = yx_tile_shape
        downsample = 0
        while True:
            new_downsample = downsample + 1
            next_downsample_shape, new_downsample_tile_shape = block_downsample_shape(
                yx_shape, new_downsample, yx_tile_shape
            )
            if not any(x > max(0, y) for x, y in zip(downsample_shape, shape)):
                break
            downsample_shape = next_downsample_shape
            downsample_tile_shape = new_downsample_tile_shape
            downsample = new_downsample
        # NOTE: Assuming channels last
        channels = self.shape[-1]
        thumbnail = np.zeros(downsample_shape + (channels,), dtype=np.uint8)
        # Resize tiles to new_downsample_tile_shape and combine
        tile_indexes = list(np.ndindex(self_mosaic_shape))
        for tile_index in self.pbar(tile_indexes, desc="Generating thumbnail"):
            tile = self[tile_slices(tile_index, yx_tile_shape)]
            tile = mean_pool(tile.astype(float), downsample).astype(np.uint8)
            thumbnail[tile_slices(tile_index, downsample_tile_shape)] = tile
        if approx_ok:
            return thumbnail
        return resize_array(thumbnail, shape, "bicubic")

    @staticmethod
    def pbar(iterable: Iterable, *args, **kwargs) -> Iterator:
        """Return an iterator that displays a progress bar.

        Uses tqdm if installed, otherwise falls back to a simple iterator.

        Args:
            iterable (Iterable):
                Iterable to iterate over.
            args (tuple):
                Positional arguments to pass to tqdm.
            kwargs (dict):
                Keyword arguments to pass to tqdm.

        Returns:
            Iterator: Iterator that displays a progress bar.
        """
        try:
            from tqdm.auto import tqdm
        except ImportError:

            def tqdm(x, *args, **kwargs):
                return x

        return tqdm(iterable, *args, **kwargs)


def get_tile(
    queue: Queue,
    ji: Tuple[int, int],
    tilesize: Tuple[int, int],
    path: Path,
) -> np.ndarray:
    """Append a tile read from a reader to a multiprocessing queue.

    Args:
        queue (Queue):
            A multiprocessing Queue to put tiles on to.
        ji (Tuple[int, int]):
            Index of tile.
        tilesize (Tuple[int, int]):
            Tile size as (width, height).
        path (Path):
            Path to file to read from.

    Returns:
        Tuple[Tuple[int, int], np.ndarray]:
            Tuple of the tile index and the tile.
    """
    reader = Reader.from_file(path)
    # Read the tile
    j, i = ji
    index = (
        slice(j * tilesize[1], (j + 1) * tilesize[1]),
        slice(i * tilesize[0], (i + 1) * tilesize[0]),
    )
    # Filter warnings (e.g. from gylmur about reading past the edge)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        tile = reader[index]
    queue.put((ji, tile))


class MultiProcessTileIterator:
    """An iterator which returns tiles generated by a reader.

    This is a fancy iterator that uses a multiprocress queue to
    accelerate the reading of tiles. It can also use an intermediate
    file to allow for reading and writing with different tile sizes.

    Args:
        reader (Reader):
            Reader for image.
        read_tile_size (Tuple[int, int]):
            Tile size to read from reader.
        yield_tile_size (Optional[Tuple[int, int]]):
            Tile size to yield. If None, yield_tile_size = read_tile_size.
        num_workers (int):
            Number of workers to use.
        intermediate (Path):
            Intermediate reader/writer to use. Must support random
            access reads and writes.
        verbose (bool):
            Verbose output.

    Yields:
        np.ndarray:
            A tile from the reader.
    """

    def __init__(
        self,
        reader: Reader,
        read_tile_size: Tuple[int, int],
        yield_tile_size: Optional[Tuple[int, int]] = None,
        num_workers: int = None,
        intermediate=None,
        verbose: bool = False,
        timeout: float = 10.0,
    ) -> None:
        self.reader = reader
        self.shape = reader.shape
        self.read_tile_size = read_tile_size
        self.yield_tile_size = yield_tile_size or read_tile_size
        self.intermediate = intermediate
        self.verbose = verbose
        self.timeout = timeout if timeout >= 0 else float("inf")
        self.processes = set()
        self.queue = Queue()
        self.enqueued = set()
        self.reordering_dict = {}
        self.read_j = 0
        self.read_i = 0
        self.yield_i = 0
        self.yield_j = 0
        self.num_workers = num_workers or os.cpu_count() or 2
        self.read_mosaic_shape = mosaic_shape(
            self.shape,
            self.read_tile_size[::-1],
        )
        self.yield_mosaic_shape = mosaic_shape(
            self.shape,
            self.yield_tile_size[::-1],
        )
        self.remaining_reads = list(np.ndindex(self.read_mosaic_shape))
        self.tile_status = zarr.zeros(self.yield_mosaic_shape, dtype="u1")
        try:
            from tqdm.auto import tqdm

            self.read_pbar = tqdm(
                total=np.prod(self.read_mosaic_shape),
                desc="Reading",
                colour="cyan",
                smoothing=0.01,
            )
        except ImportError:
            self.read_pbar = None

        # Validation and error handling
        if self.read_tile_size != self.yield_tile_size and not self.intermediate:
            raise ValueError(
                f"read_tile_size ({self.read_tile_size})"
                f" != yield_tile_size ({self.yield_tile_size})"
                " and intermediate is not set. An intermediate is"
                " required when the read and yield tile size differ."
            )

    def __len__(self) -> int:
        """Return the number of tiles in the reader."""
        return int(np.prod(self.yield_mosaic_shape))

    def __iter__(self) -> Iterator:
        """Return an iterator for the reader."""
        self.read_j = 0
        self.read_i = 0
        return self

    @property
    def read_index(self) -> Tuple[int, int]:
        """Return the current read index."""
        return self.read_j, self.read_i

    @read_index.setter
    def read_index(self, value: Tuple[int, int]) -> None:
        """Set the current read index."""
        self.read_j, self.read_i = value

    @property
    def yield_index(self) -> Tuple[int, int]:
        """Return the current yield index."""
        return self.yield_j, self.yield_i

    @yield_index.setter
    def yield_index(self, value: Tuple[int, int]) -> None:
        """Set the current yield index."""
        self.yield_j, self.yield_i = value

    def wrap_indexes(self) -> None:
        """Wrap the read and yield indexes."""
        self.read_index, overflow = wrap_index(self.read_index, self.read_mosaic_shape)
        if overflow and self.verbose:
            print("All tiles read.")
        self.yield_index, overflow = wrap_index(
            self.yield_index, self.yield_mosaic_shape
        )
        if overflow:
            if self.verbose:
                print("All tiles yielded.")
            self.close()
            raise StopIteration

    def __next__(self) -> np.ndarray:
        """Return the next tile from the reader."""
        # Ensure a valid read ij index
        self.wrap_indexes()

        # Add tile reads to the queue until the maximum number of
        # workers is reached
        self.fill_queue()

        # Get the next yield tile from the queue
        t0 = time.perf_counter()
        while (time.perf_counter() - t0) < self.timeout:
            # Remove all tiles from the queue into the reordering dict
            self.empty_queue()

            # Remove the next read tile from the reordering dict. May be
            # None if the read tile is not in the reordering dict or if
            # an intermediate is used.
            tile = self.pop_next_read_tile()

            # Return the tile if no intermediate is being used and the
            # tile was in the reordering dict.
            if not self.intermediate and tile is not None:
                return tile

            # Get the next tile from the intermediate. Returns None if
            # intermediate is None or the tile is not in the
            # intermediate.
            tile = self.read_next_from_intermediate()
            if tile is not None:
                return tile

            # Ensure the queue is kept full
            self.fill_queue()

            # Sleep and try again
            time.sleep(0.1)
        warnings.warn(
            "Failed to get next tile after 100 attempts. Dumping debug information."
        )
        print(f"Reader Shape {self.reader.shape}")
        print(f"Read Tile Size {self.read_tile_size}")
        print(f"Yield Tile Size {self.yield_tile_size}")
        print(f"Read Mosaic Shape {self.read_mosaic_shape}")
        print(f"Yield Mosaic Shape {self.yield_mosaic_shape}")
        print(f"Read Index {self.read_index}")
        print(f"Yield Index {self.yield_index}")
        print(f"Remaining Reads (:10) {self.remaining_reads[:10]}")
        print(f"Enqueued {self.enqueued}")
        print(f"Reordering Dict (keys) {self.reordering_dict.keys()}")
        print(f"Queue Size {self.queue.qsize()}")
        intermediate_read_slices = tile_slices(
            index=(self.yield_j, self.yield_i),
            shape=self.yield_tile_size[::-1],
        )
        print(f"Intermediate Read slices {intermediate_read_slices}")
        # Terminate the read processes
        self.close()
        raise IOError(f"Tile read timed out at index {self.yield_index}")

    def read_next_from_intermediate(self) -> Optional[np.ndarray]:
        """Read the next tile from the intermediate file."""
        if self.intermediate is None:
            return None
        intermediate_read_slices = tile_slices(
            index=(self.yield_j, self.yield_i),
            shape=self.yield_tile_size[::-1],
        )
        status = self.tile_status[self.yield_index]
        if np.all(status == 1):  # Intermediate has all data for the tile
            self.tile_status[self.yield_index] = 2
            self.yield_i += 1
            return self.intermediate[intermediate_read_slices]
        return None

    def empty_queue(self) -> None:
        """Remove all tiles from the queue into the reordering dict."""
        while not self.queue.empty():
            ji, tile = self.queue.get()
            self.reordering_dict[ji] = tile

    def fill_queue(self) -> None:
        """Add tile reads to the queue until the max number of workers is reached."""
        while len(self.enqueued) < self.num_workers and len(self.remaining_reads) > 0:
            next_ji = self.remaining_reads.pop(0)
            process = multiprocessing.Process(
                target=get_tile,
                args=(
                    self.queue,
                    next_ji,
                    self.read_tile_size,
                    self.reader.path,
                ),
                daemon=True,
            )
            process.start()
            self.processes.add(process)
            self.enqueued.add(next_ji)

    def update_read_pbar(self) -> None:
        """Update the read progress bar."""
        if self.read_pbar is not None:
            self.read_pbar.update()

    def pop_next_read_tile(self) -> Optional[np.ndarray]:
        """Remove the next tile from the reordering dict.

        Returns:
            Optional[np.ndarray]:
                The next tile from the reordering dict if available.
                If an intermediate is being used, or the tile is not
                in the reordering dict, this will be None.
        """
        read_ji = (self.read_j, self.read_i)
        if read_ji in self.reordering_dict:
            self.enqueued.remove(read_ji)
            tile = self.reordering_dict.pop(read_ji)

            # If no intermediate is required, return the tile
            if not self.intermediate:
                if tile is None:
                    raise Exception(f"Tile {read_ji} is None")
                self.read_i += 1
                self.update_read_pbar()
                self.yield_i += 1
                return tile

            # Otherwise, write the tile to the intermediate
            intermediate_write_index = tile_slices(
                index=read_ji,
                shape=self.read_tile_size,
            )
            self.intermediate[intermediate_write_index] = tile
            tile_status_index = tuple(
                slice(max(0, floor(x.start / r)), ceil(x.stop / r))
                for x, r in zip(intermediate_write_index, self.yield_tile_size)
            )
            self.tile_status[tile_status_index] = 1
            self.read_i += 1
            self.update_read_pbar()
        return None

    def close(self):
        """Safely end any dependants (threads, processes, and files).

        Close progress bars and join child processes. Terminate children
        if they fail to join after one second.
        """
        if self.read_pbar is not None:
            self.read_pbar.close()
        # Join processes in parallel threads
        with ThreadPoolExecutor(len(self.processes)) as executor:
            executor.map(lambda p: p.join(1), self.processes)
        # Terminate any child processes if still alive
        for process in self.processes:
            if process.is_alive():
                process.terminate()

    def __del__(self):
        """Destructor."""
        self.close()


class JP2Reader(Reader):
    """Reader for JP2 files using glymur.

    Args:
        path (Path): Path to file.
    """

    def __init__(self, path: Path) -> None:
        super().__init__(path)
        import glymur

        # Enable multithreading
        if glymur.options.version.openjpeg_version_tuple >= (2, 2, 0):
            glymur.set_option("lib.num_threads", multiprocessing.cpu_count())

        self.jp2 = glymur.Jp2k(str(path))
        self.shape = self.jp2.shape
        self.dtype = np.uint8
        self.axes = "YXS"
        self.microns_per_pixel = self._get_mpp()
        self.tile_shape = self._get_tile_shape()

    def get_tile(self, index: Tuple[int, int], decode: bool = True) -> np.ndarray:
        """Get tile at index.

        Args:
            index (Tuple[int, int]):
                The index of the tile to get.
            decode (bool, optional):
                Whether to decode the tile. Defaults to True.

        Returns:
            np.ndarray:
                The tile at index.
        """
        if not decode:
            raise NotImplementedError("Returning encoded JP2 tiles is not supported.")
        return self.jp2[tile_slices(index, self.tile_shape)]

    def _get_mpp(self) -> Optional[Tuple[float, float]]:
        """Get the microns per pixel for the image.

        Returns:
            Optional[Tuple[float, float]]:
                The resolution of the image in microns per pixel.
                If the resolution is not available, this will be None.
        """
        import glymur

        boxes = {type(box): box for box in self.jp2.box}
        header_box = boxes[glymur.jp2box.JP2HeaderBox]
        header_sub_boxes = {type(box): box for box in header_box.box}
        resolution_box = header_sub_boxes.get(glymur.jp2box.ResolutionBox)
        if resolution_box is None:
            return None
        resolution_sub_boxes = {type(box): box for box in resolution_box.box}
        capture_resolution_box = resolution_sub_boxes.get(
            glymur.jp2box.CaptureResolutionBox
        )
        if capture_resolution_box is None:
            return None
        y_res = capture_resolution_box.vertical_resolution
        x_res = capture_resolution_box.horizontal_resolution
        return ppu2mpp(x_res, "cm"), ppu2mpp(y_res, "cm")

    def _get_tile_shape(self) -> Tuple[int, int]:
        """Get the tile shape as a (height, width) tuple.

        Returns:
            Tuple[int, int]:
                The tile shape as (height, width).
        """
        import glymur

        boxes = {type(box): box for box in self.jp2.box}
        ccb = boxes.get(glymur.jp2box.ContiguousCodestreamBox, None)
        if ccb is None:
            raise ValueError("No codestream box found.")
        segments = {type(segment): segment for segment in ccb.codestream.segment}
        siz = segments.get(glymur.codestream.SIZsegment, None)
        if siz is None:
            raise ValueError("No SIZ segment found.")
        return (siz.ytsiz, siz.xtsiz)

    def __getitem__(self, index: tuple) -> np.ndarray:
        """Get pixel data at index."""
        return self.jp2[index]

    def thumbnail(self, shape: Tuple[int, ...], approx_ok: bool = False) -> np.ndarray:
        scale = scale_to_fit(self.shape[:2], shape)
        downsample = 1 / scale
        out_shape = tuple(int(x * scale) for x in self.shape[:2])
        # Glymur requires a power of two stride
        pow_2_downsample = 2 ** np.floor(np.log2(downsample))
        # Get the power of two downsample
        thumbnail = self.jp2[::pow_2_downsample, ::pow_2_downsample]
        # Resize the thumbnail if required
        if approx_ok:
            return thumbnail
        return resize_array(thumbnail, out_shape)


class TIFFReader(Reader):
    """Reader for TIFF files using tifffile."""

    def __init__(self, path: Path) -> None:
        """Initialize reader.

        Args:
            path (Path): Path to file.
        """
        import tifffile

        super().__init__(path)
        self.tiff = tifffile.TiffFile(str(path))
        self.tiff_page = self.tiff.pages[0]
        self.microns_per_pixel = self._get_mpp()
        self.array = self.tiff_page.asarray()
        self.shape = self.array.shape
        self.dtype = self.array.dtype
        self.axes = self.tiff.series[0].axes
        self.is_tiled = self.tiff_page.is_tiled
        self.tile_shape = None
        self.mosaic_shape = None
        self.mosaic_byte_offsets = None
        self.mosaic_byte_counts = None
        if self.is_tiled:
            self.tile_shape = (self.tiff_page.tilelength, self.tiff_page.tilewidth)
            self.mosaic_shape = mosaic_shape(
                array_shape=self.shape, tile_shape=self.tile_shape
            )
            self.mosaic_byte_offsets = np.array(self.tiff_page.dataoffsets).reshape(
                self.mosaic_shape
            )
            self.mosaic_byte_counts = np.array(self.tiff_page.databytecounts).reshape(
                self.mosaic_shape
            )
        self.jpeg_tables = self.tiff_page.jpegtables
        self.color_space: ColorSpace = ColorSpace.from_tiff(self.tiff_page.photometric)
        self.codec: Codec = Codec.from_tiff(self.tiff_page.compression)
        self.compression_level = None  # To be filled in if known later

    def _get_mpp(self) -> Optional[Tuple[float, float]]:
        """Get the microns per pixel for the image.

        This checks the resolution and resolution unit TIFF tags.

        Returns:
            Optional[Tuple[float, float]]:
                The resolution of the image in microns per pixel.
                If the resolution is not available, this will be None.
        """
        try:
            y_resolution = self.tiff_page.tags["YResolution"].value[0]
            x_resolution = self.tiff_page.tags["XResolution"].value[0]
            resolution_units = self.tiff_page.tags["ResolutionUnit"].value
            return ppu2mpp(x_resolution, resolution_units), ppu2mpp(
                y_resolution, resolution_units
            )
        except KeyError:
            return None

    def get_tile(self, index: Tuple[int, int], decode: bool = True) -> np.ndarray:
        """Get tile at index.

        Args:
            index (Tuple[int, int]):
                The index of the tile to get.
            decode (bool, optional):
                Whether to decode the tile. Defaults to True.

        Returns:
            np.ndarray:
                The tile at index.
        """
        flat_index = index[0] * self.tile_shape[1] + index[1]
        fh = self.tiff.filehandle
        _ = fh.seek(self.mosaic_byte_offsets[index])
        data = fh.read(self.mosaic_byte_counts[index])
        if not decode:
            return data
        tile, _, _ = self.tiff_page.decode(
            data, flat_index, jpegtables=self.tiff_page.jpegtables
        )
        return tile

    def __getitem__(self, index: Tuple[Union[slice, int]]) -> np.ndarray:
        """Get pixel data at index."""
        return self.array[index]


class DICOMWSIReader(Reader):
    """Reader for DICOM Whole Slide Images (WSIs) using wsidicom.

    DICOM Whole Slide Imaging:  https://dicom.nema.org/Dicom/DICOMWSI/
    """

    def __init__(self, path: Path) -> None:
        """Initialize reader.

        Args:
            path (Path):
                Path to file.
        """
        from wsidicom import WsiDicom

        super().__init__(path)
        self.slide = WsiDicom.open(self.path)
        channels = len(self.slide.read_tile(0, (0, 0)).getbands())
        self.shape = (self.slide.size.height, self.slide.size.width, channels)
        self.dtype = np.uint8
        self.microns_per_pixel = (
            self.slide.base_level.mpp.height,
            self.slide.base_level.mpp.width,
        )
        self.tile_shape = (self.slide.tile_size.height, self.slide.tile_size.width)
        self.mosaic_shape = mosaic_shape(self.shape, self.tile_shape)
        dataset = self.slide.base_level.datasets[0]
        # Sanity check
        if np.prod(self.mosaic_shape) != int(dataset.NumberOfFrames):
            raise ValueError(
                "Number of frames in DICOM dataset does not match mosaic shape."
            )
        self.codec: Codec = Codec.from_string(dataset.LossyImageCompressionMethod)
        self.compression_level = (
            None  # Set if known: dataset.get(LossyImageCompressionRatio)?
        )
        self.color_space = ColorSpace.from_dicom(dataset.photometric_interpretation)
        self.jpeg_tables = None

    def get_tile(self, index: Tuple[int, int], decode: bool = True) -> np.ndarray:
        """Get tile at index.

        Args:
            index (Tuple[int, int]):
                The index of the tile to get.
            decode (bool, optional):
                Whether to decode the tile. Defaults to True.

        Returns:
            np.ndarray:
                The tile at index.
        """
        if decode:
            return np.array(self.slide.read_tile(level=0, tile=index[::-1], z=0))
        return self.slide.read_encoded_tile(level=0, tile=index[::-1], z=0)

    def __getitem__(self, index: Tuple[Union[slice, int]]) -> np.ndarray:
        """Get pixel data at index."""
        if index is ...:
            return np.array(self.slide.base_level.get_default_full())
        xs = index[1]
        ys = index[0]
        start_x = xs.start or 0
        start_y = ys.start or 0
        end_x = xs.stop or self.shape[1]
        end_y = ys.stop or self.shape[0]

        # Prevent reading past the edges of the image
        end_x = min(end_x, self.shape[1])
        end_y = min(end_y, self.shape[0])

        # Read the image
        img = self.slide.read_region(
            location=(start_x, start_y),
            level=0,
            size=(end_x - start_x, end_y - start_y),
            z=0,
        )
        return np.array(img.convert("RGB"))


class OpenSlideReader(Reader):
    """Reader for OpenSlide files using openslide-python."""

    def __init__(self, path: Path) -> None:
        import openslide

        super().__init__(path)
        self.os_slide = openslide.OpenSlide(str(path))
        self.shape = self.os_slide.level_dimensions[0][::-1] + (3,)
        self.dtype = np.uint8
        self.axes = "YXS"
        self.tile_shape = None  # No easy way to get tile shape currently
        self.microns_per_pixel = self._get_mpp()

    def get_tile(self, index: Tuple[int, int], decode: bool = True) -> np.ndarray:
        """Get tile at index.

        Args:
            index (Tuple[int, int]):
                The index of the tile to get.
            decode (bool, optional):
                Whether to decode the tile. Defaults to True.

        Returns:
            np.ndarray:
                The tile at index.
        """
        raise NotImplementedError("OpenSlideReader does not support reading tiles.")

    def _get_mpp(self) -> Optional[Tuple[float, float]]:
        """Get the microns per pixel for the image.

        Returns:
            Optional[Tuple[float, float]]:
                The microns per pixel as (x, y) tuple.
        """
        try:
            return (
                self.os_slide.properties["openslide.mpp-x"],
                self.os_slide.properties["openslide.mpp-y"],
            )
        except KeyError:
            warnings.warn("OpenSlide could not find MPP.")
        # Fall back to TIFF resolution tags
        try:
            resolution = (
                float(self.os_slide.properties["tiff.XResolution"]),
                float(self.os_slide.properties["tiff.YResolution"]),
            )
            units = self.os_slide.properties["tiff.ResolutionUnit"]
            self._check_sensible_resolution(resolution, units)
            return tuple(ppu2mpp(x, units) for x in resolution)
        except KeyError:
            warnings.warn("No resolution metadata found.")
        return None

    @staticmethod
    def _check_sensible_resolution(
        tiff_resolution: Tuple[float, float], tiff_units: int
    ) -> None:
        """Check whether the resolution is sensible.

        It is common for TIFF files to have incorrect resolution tags.
        This method checks whether the resolution is sensible and warns
        if it is not.

        Args:
            tiff_resolution (Tuple[float, float]):
                The TIFF resolution as an (x, y) tuple.
            tiff_units (int):
                The TIFF units of the resolution. A value of 2 indicates
                inches and a value of 3 indicates centimeters.
        """
        if tiff_units == 2 and 72 in tiff_resolution:
            warnings.warn(
                "TIFF resolution tags found."
                " However, they have a common default value of 72 pixels per inch."
                " This may from a misconfigured software library or tool"
                " which is expecting to handle print documents."
            )
        if 0 in tiff_resolution:
            warnings.warn(
                "TIFF resolution tags found."
                " However, one or more of the values is zero."
            )

    def __getitem__(self, index: Tuple[Union[int, slice], ...]) -> np.ndarray:
        """Get pixel data at index."""
        xs = index[1]
        ys = index[0]
        start_x = xs.start or 0
        start_y = ys.start or 0
        end_x = xs.stop or self.shape[1]
        end_y = ys.stop or self.shape[0]

        # Prevent reading past the edges of the image
        end_x = min(end_x, self.shape[1])
        end_y = min(end_y, self.shape[0])

        # Read the image
        img = self.os_slide.read_region(
            location=(start_x, start_y),
            level=0,
            size=(end_x - start_x, end_y - start_y),
        )
        return np.array(img.convert("RGB"))

    def thumbnail(self, shape: Tuple[int, ...], approx_ok: bool = False) -> np.ndarray:
        warn_unused(approx_ok, ignore_falsey=True)
        return np.array(self.os_slide.get_thumbnail(shape[::-1]))



import dataclasses
import multiprocessing
import shutil
import tempfile
import uuid
import warnings
from abc import ABC, abstractmethod
from functools import partial
from math import floor
from pathlib import Path
from typing import Any, Callable, Iterable, Iterator, List, Optional, Tuple, Union

import numpy as np
import zarr

from wsic import __version__ as wsic_version
from wsic.codecs import register_codecs
from wsic.enums import Codec, ColorSpace
from wsic.metadata import ngff
from wsic.readers import DICOMWSIReader, MultiProcessTileIterator, Reader, TIFFReader
from wsic.types import PathLike
from wsic.utils import (
    downsample_shape,
    mean_pool,
    mosaic_shape,
    mpp2ppu,
    scale_to_fit,
    tile_slices,
    warn_unused,
)


class Writer(ABC):
    """Base class for image writers.

    Args:
        path (PathLike):
            Path to the output file.
        shape (Tuple[int, int]):
            Shape of the output image.
        tile_size (Tuple[int, int], optional):
            A (width, height) tuple of output tile size in pixels.
            Defaults to (256, 256).
        dtype (np.dtype, optional):
            Data type of the output image. Defaults to np.uint8.
        color_space (ColorSpace, optional):
            Color space the output image.
            Defaults to "rgb".
        compression (str, optional):
            Compression codec to use. Defaults to None. Not all
            writers support compression.
        compression_level (int, optional):
            Compression level to use. Defaults to 0 (lossless /
            maximum).
        microns_per_pixel (Tuple[float, float], optional):
            A (width, height) tuple of microns per pixel. Defaults to
            None.
        pyramid_downsamples (List[int], optional):
            A list of downsamples to use in the pyramid. Defaults to
            None. Not all writers support pyramids.
        overwrite (bool, optional):
            Overwrite output file if it exists.
            Defaults to False.
        verbose (bool, optional):
            Print more output. Defaults to False.
    """

    def __init__(
        self,
        path: PathLike,
        shape: Tuple[int, int],
        tile_size: Tuple[int, int] = (256, 256),
        dtype: np.dtype = np.uint8,
        color_space: Optional[ColorSpace] = ColorSpace.RGB,
        codec: Optional[Codec] = None,
        compression_level: int = 0,
        microns_per_pixel: Tuple[float, float] = None,
        pyramid_downsamples: Optional[List[int]] = None,
        overwrite: bool = False,
        verbose: bool = False,
    ):
        self.path = Path(path)
        self.shape = shape
        self.tile_size = tile_size
        self.dtype = dtype
        self.color_space = color_space or "rgb"
        self.codec = Codec.from_string(codec) if codec else None
        self.compression_level = compression_level or 0
        self.microns_per_pixel = microns_per_pixel
        self.pyramid_downsamples = (
            list(pyramid_downsamples) if pyramid_downsamples else []
        )
        self.overwrite = overwrite
        self.verbose = verbose

        if self.path.exists() and not self.overwrite:
            raise FileExistsError(f"{self.path} already exists")

    def reader_tile_iterator(
        self,
        reader: Reader,
        num_workers: int = 2,
        read_tile_size: Tuple[int, int] = None,
        yield_tile_size: Tuple[int, int] = None,
        intermediate: zarr.Group = None,
        timeout: float = 10.0,
    ) -> Iterator:
        """Returns an iterator which returns tiles generated by reader.

        Args:
            reader (Reader):
                Reader to read tiles from.
            num_workers (int, optional):
                Number of workers to use. Defaults to 2.
            read_tile_size (Tuple[int, int], optional):
                A (width, height) tuple of read tile size in pixels.
                Defaults to self.tile_size.
            intermediate (np.ndarray, optional):
                An intermediate image to write tiles to.

        Returns:
            Iterator: Iterator which returns tiles generated by reader.

        """
        if read_tile_size is None:
            read_tile_size = self.tile_size
        return MultiProcessTileIterator(
            reader=reader,
            read_tile_size=read_tile_size,
            yield_tile_size=yield_tile_size or self.tile_size,
            intermediate=intermediate,
            num_workers=num_workers,
            verbose=self.verbose,
            timeout=timeout,
        )

    def __setitem__(
        self, index: Tuple[Union[int, slice], ...], value: np.ndarray
    ) -> None:
        """Return pixel data at index."""
        raise NotImplementedError()

    @abstractmethod
    def copy_from_reader(
        self,
        reader: Reader,
        num_workers: int = 2,
        read_tile_size: Tuple[int, int] = None,
        timeout: float = 10.0,
    ) -> None:
        """Write pixel data to by copying from a Reader.

        Args:
            reader (Reader):
                Reader object.
            num_workers (int, optional):
                Number of workers to use. Defaults to 2.
            read_tile_size (Tuple[int, int], optional):
                Tile size to read. Defaults to None.
                This will use the tile size of the writer if None.
            timeout (float, optional):
                Timeout for workers. Defaults to 10s.
        """
        if self.path.exists() and not self.overwrite:
            raise FileExistsError(f"{self.path} exists and overwrite is False.")

    @staticmethod
    def level_progress(iterable: Iterable, **kwargs) -> Iterator:
        """Wrapper for a tile yeilding iterable when writing a level.

        Used to display progress when copying from a reader.

        Some of the tqdm defaults are overridden but can be changed by
        passing values as kwargs. Parameters which differ to the tqdm
        defaults here are:
        - `smoothing = 0.1`
        - `colour = "magenta"`

        Args:
            iterable (Iterable):
                The iterable to wrap.
            **kwargs (dict):
                Extra kwargs for tqdm. Overrides defaults.
        """
        tqdm_kwargs = {
            "colour": "magenta",
            "smoothing": 0.01,
            "desc": "Writing",
        }
        tqdm_kwargs.update(kwargs)
        try:
            from tqdm.auto import tqdm

            return tqdm(iterable, **tqdm_kwargs)
        except ImportError:
            return iterable

    @staticmethod
    def pyramid_progress(iterable: Iterable, **kwargs) -> Iterator:
        """Wrap an iterable in a progress bar.

        Used to display progress when copying from a reader.

        Some of the tqdm defaults are overridden but can be changed by
        passing values as kwargs. Parameters which differ to the tqdm
        defaults here are:
        - `smoothing = 0`
        - `colour = "magenta"`

        Args:
            iterable (Iterable):
                The iterable to wrap.
            **kwargs (dict):
                Extra kwargs for tqdm. Overrides defaults.
        """
        tqdm_kwargs = {
            "colour": "blue",
            # Bar format with no ETA
            "bar_format": "{l_bar}{bar}| {n_fmt}/{total_fmt}",
            "desc": "Building Pyramid",
        }
        tqdm_kwargs.update(kwargs)
        try:
            from tqdm.auto import tqdm

            return tqdm(iterable, **tqdm_kwargs)
        except ImportError:
            return iterable

    @staticmethod
    def transcode_progress(iterable: Iterable, **kwargs) -> Iterable:
        """Progress bar for transcoding.

        Args:
            iterable (Iterable):
                Iterable to wrap.
            **kwargs:
                Keyword arguments to pass to the progress bar.

        Returns:
            Iterable:
        """
        try:
            from tqdm.auto import tqdm

            return tqdm(
                iterable,
                desc="Transcoding",
                colour="green",
                **kwargs,
            )
        except ImportError:
            return iterable


class JP2Writer(Writer):
    """Tile-wise JP2 writer using glymur.

    Note that when writing tiled JP2 files, the tiles must all be the
    same size and must be written in the order left-to-right, then
    top-to-bottom (row-by-row). Tiles cannot be skipped.

    Args:
        path (PathLike):
            Path to output file.
        shape (Tuple[int, int]):
            A (width, height) tuple of image size in pixels.
        tile_size (Tuple[int, int], optional):
            A (width, height) tuple of tile size in pixels.
            Defaults to (256, 256).
        dtype (np.dtype, optional):
            Data type of output image. Defaults to np.uint8.
        color_space (ColorSpace, optional):
            Color space the output image.
            Defaults to "rgb".
        compression (str, optional):
            Compression type. Currently only JPEG 2000 compression is
            supported. Defaults to None.
        compression_level (int, optional):
            Compression level. Currently unused. Defaults to None.
        microns_per_pixel (Tuple[float, float], optional):
            A (width, height) tuple of microns per pixel.
            Defaults to None.
        pyramid_downsamples (List[int], optional):
            A list of downsamples to create. Unused but included
            for API consistency. Defaults to None.
        overwrite (bool, optional):
            Overwrite existing file. Defaults to False.
        verbose (bool, optional):
            Print more output. Defaults to False.

    """

    def __init__(
        self,
        path: PathLike,
        shape: Tuple[int, int],
        tile_size: Tuple[int, int] = (256, 256),
        dtype: np.dtype = np.uint8,
        color_space: Optional[ColorSpace] = "rgb",  # Currently unused
        codec: Optional[Codec] = "jpeg2000",  # Currently unused
        compression_level: int = 0,  # Currently unused
        microns_per_pixel: Optional[Tuple[float, float]] = None,  # Currently unused
        pyramid_downsamples: Optional[List[int]] = None,  # Unused
        overwrite: bool = False,
        verbose: bool = False,
    ) -> None:
        if color_space != "rgb":
            warn_unused(color_space)
        if codec != "jpeg2000":
            warn_unused(codec)
        warn_unused(compression_level, ignore_falsey=True)
        warn_unused(microns_per_pixel)
        warn_unused(pyramid_downsamples, ignore_falsey=True)
        super().__init__(
            path=path,
            shape=shape,
            tile_size=tile_size,
            dtype=dtype,
            color_space=color_space,
            codec=codec,
            compression_level=compression_level,
            microns_per_pixel=microns_per_pixel,
            pyramid_downsamples=pyramid_downsamples,
            overwrite=overwrite,
            verbose=verbose,
        )

    def __setitem__(self, index: Tuple[int, ...], value: np.ndarray) -> None:
        """Write pixel data at index. Not supported for JP2Writer."""
        raise NotImplementedError("JP2 files do not support random access writes.")

    def copy_from_reader(
        self,
        reader: Reader,
        num_workers: int = 2,
        read_tile_size: Optional[Tuple[int, int]] = None,
        timeout: float = 10.0,
    ) -> None:
        """Write pixel data to by copying from a Reader.

        Args:
            reader (Reader):
                Reader object.
            num_workers (int, optional):
                Number of workers to use. Defaults to 2.
            read_tile_size (Tuple[int, int], optional):
                Tile size to read. Defaults to None.
                This will use the tile size of the writer if None.
            timeout (float, optional):
                Timeout for workers. Defaults to 10s.
        """
        super().copy_from_reader(
            reader=reader,
            num_workers=num_workers,
            read_tile_size=read_tile_size,
            timeout=timeout,
        )
        import glymur

        jp2 = glymur.Jp2k(
            self.path, shape=reader.shape, tilesize=self.tile_size, verbose=self.verbose
        )
        reader_tile_iterator = self.reader_tile_iterator(
            reader=reader,
            num_workers=num_workers,
            read_tile_size=read_tile_size or self.tile_size,
            timeout=timeout,
        )
        reader_tile_iterator = iter(self.level_progress(reader_tile_iterator))
        for tile_writer in jp2.get_tilewriters():
            try:
                tile_writer[:] = next(reader_tile_iterator)
            except StopIteration:
                raise StopIteration(
                    "Reader tile iterator stopped early. "
                    "Glymur is expecting more tiles to be written."
                )


class TIFFWriter(Writer):
    """Tile-wise TIFF writer using tifffile.

    Note that when writing tiled TIFF files, the tiles must all be the
    same size and must be written in the order left-to-right, then
    top-to-bottom (row-by-row). Tiles cannot be skipped.

    TIFF 6.0 Specification Notes
    ----------------------------

    - TileWidth and TileLength (height) must each be a multiple of 16.
    - "Offsets [bytes from the start of file to each tile blob and
      therefore the tile ordering when writing] are ordered
      left-to-right and top-to-bottom."
    - "For PlanarConfiguration = 2, the offsets for the first component
      plane are stored first, followed by all the offsets for the second
      component plane, and so on."

      Full specification at:
      https://web.archive.org/web/20210108174645/https://www.adobe.io/content/dam/udp/en/open/standards/tiff/TIFF6.pdf


    Args:
        path (PathLike):
            Path to output file.
        shape (Tuple[int, int]):
            A (width, height) tuple of image size in pixels.
        tile_size (Tuple[int, int], optional):
            A (width, height) tuple of tile size in pixels.
            Defaults to (256, 256).
        dtype (np.dtype, optional):
            Data type of output image. Defaults to np.uint8.
        color_space (ColorSpace, optional):
            color_space. Defaults to "rgb".
        compression (str, optional):
            Compression type.
            Defaults to "jpeg".
        compression_level (int, optional):
            Compression level. Defaults to -1 (highest / lossless).
        microns_per_pixel (Tuple[float, float], optional):
            A (width, height) tuple of microns per pixel.
            Defaults to None.
        pyramid_downsamples (List[int], optional):
            A list of downsamples to create. Should be strictly
            inceasing for maximum compatibility.
            Defaults to None.
        overwrite (bool, optional):
            Overwrite existing file. Defaults to False.
        verbose (bool, optional):
            Print more output. Defaults to False.
        ome (bool):
            Write OME-TIFF metadata. Defaults to False.
    """

    def __init__(
        self,
        path: Path,
        shape: Tuple[int, int],
        tile_size: Tuple[int, int] = (256, 256),
        dtype: np.dtype = np.uint8,  # Currently unused
        color_space: Optional[ColorSpace] = "rgb",
        codec: Optional[Codec] = "jpeg",
        compression_level: int = -1,  # Currently unused
        microns_per_pixel: Tuple[float, float] = None,
        pyramid_downsamples: Optional[List[int]] = None,
        overwrite: bool = False,
        verbose: bool = False,
        *,
        ome: bool = True,
    ) -> None:
        if dtype is not np.uint8:
            warn_unused(dtype)
        super().__init__(
            path=path,
            shape=shape,
            tile_size=tile_size,
            dtype=dtype,
            color_space=color_space,
            codec=codec,
            compression_level=compression_level,
            microns_per_pixel=microns_per_pixel,
            pyramid_downsamples=pyramid_downsamples,
            overwrite=overwrite,
            verbose=verbose,
        )
        self.ome = ome

    def __setitem__(self, index: Tuple[int, ...], value: np.ndarray) -> None:
        """Write pixel data at index. Not supported for TIFFWriter.

        In theory this is possible but it can be complex. If the new tile
        is larger in bytes, the tile will have to be added to the end of the
        file. The old tile will remain in the file and waste space unless it
        is later overwritten by another of length smaller or equal to the
        original tile.
        """
        raise NotImplementedError(
            "Compressed tiled TIFF files do not support random access writes."
        )

    def copy_from_reader(
        self,
        reader: Reader,
        num_workers: int = 2,
        read_tile_size: Optional[Tuple[int, int]] = None,
        timeout: float = 10.0,
        downsample_method: Optional[str] = None,
    ) -> None:
        """Write pixel data to by copying from a Reader.

        Args:
            reader (Reader):
                Reader object.
            num_workers (int, optional):
                Number of workers to use. Defaults to 2.
            read_tile_size (Tuple[int, int], optional):
                Tile size to read. Defaults to None.
                This will use the tile size of the writer if None.
            timeout (float, optional):
                Timeout for workers. Defaults to 10s.
            downsample_method (str, optional):
                Downsample method to use. Defaults to None.
                Valid downsample methods are: "cv2", "scipy", "np", None.
        """
        super().copy_from_reader(
            reader=reader,
            num_workers=num_workers,
            read_tile_size=read_tile_size,
            timeout=timeout,
        )
        import tifffile

        microns_per_pixel = self.microns_per_pixel or reader.microns_per_pixel
        resolution = (
            (
                round(mpp2ppu(microns_per_pixel[0], "cm")),
                round(mpp2ppu(microns_per_pixel[1], "cm")),
                "CENTIMETER",
            )
            if microns_per_pixel
            else None
        )

        with ZarrIntermediate(
            None, reader.shape, zero_after_read=False
        ) as intermediate:
            reader_tile_iterator = self.reader_tile_iterator(
                reader=reader,
                num_workers=num_workers,
                intermediate=intermediate,
                read_tile_size=read_tile_size or self.tile_size,
                timeout=timeout,
            )
            reader_tile_iterator = self.level_progress(reader_tile_iterator)
            # Write baseline (level 0)
            with tifffile.TiffWriter(
                self.path,
                bigtiff=True,
                ome=self.ome,
            ) as tif:

                metadata = {}
                if self.ome and self.microns_per_pixel:
                    metadata["PhysicalSizeXUnit"] = "µm"
                    metadata["PhysicalSizeYUnit"] = "µm"
                    metadata["PhysicalSizeX"] = self.microns_per_pixel[0]
                    metadata["PhysicalSizeY"] = self.microns_per_pixel[1]

                tif.write(
                    data=iter(reader_tile_iterator),
                    tile=self.tile_size,
                    shape=reader.shape,
                    dtype=reader.dtype,
                    photometric=self.color_space,
                    compression=(self.codec.condensed(), self.compression_level),
                    resolution=resolution,
                    subifds=len(self.pyramid_downsamples),
                    metadata=metadata,
                )
                # Write pyramid resolutions
                with multiprocessing.Pool(num_workers) as pool:
                    for level, downsample in self.pyramid_progress(
                        enumerate(self.pyramid_downsamples),
                        total=len(self.pyramid_downsamples),
                    ):
                        level_shape = tuple(
                            floor(s / downsample) for s in reader.shape[:2]
                        ) + (reader.shape[-1],)

                        level_tiles_shape = mosaic_shape(
                            level_shape,
                            self.tile_size,
                        )

                        func = partial(
                            get_level_tile,
                            tile_size=self.tile_size,
                            downsample=downsample,
                            read_intermediate_path=intermediate.path,
                            downsample_method=downsample_method,
                        )

                        tile_generator = pool.imap(
                            func=func,
                            iterable=np.ndindex(level_tiles_shape),
                        )

                        tile_generator = self.level_progress(
                            tile_generator,
                            total=int(np.product(level_tiles_shape)),
                            desc=f"Level {level}",
                            leave=False,
                        )

                        tif.write(
                            data=iter(tile_generator),
                            tile=self.tile_size,
                            shape=level_shape,
                            dtype=reader.dtype,
                            photometric=self.color_space,
                            compression=(
                                self.codec.condensed(),
                                self.compression_level,
                            ),
                            subfiletype=1,  # Subfile type: reduced resolution
                        )


class SVSWriter(Writer):
    """Aperio SVS writer using tifffile.

    Note that when writing tiled TIFF files, the tiles must all be the
    same size and must be written in the order left-to-right, then
    top-to-bottom (row-by-row). Tiles cannot be skipped.

    Args:
        path (PathLike):
            Path to output file.
        shape (Tuple[int, int]):
            A (width, height) tuple of image size in pixels.
        tile_size (Tuple[int, int], optional):
            A (width, height) tuple of tile size in pixels.
            Defaults to (256, 256).
        dtype (np.dtype, optional):
            Data type of output image. Defaults to np.uint8.
        color_space (ColorSpace, optional):
            color_space. Defaults to "rgb".
        compression (str, optional):
            Compression type.
            Defaults to "jpeg".
        compression_level (int, optional):
            Compression level. Defaults to 95. Currently unused.
        microns_per_pixel (Tuple[float, float], optional):
            A (width, height) tuple of microns per pixel.
            Defaults to None.
        pyramid_downsamples (List[int], optional):
            A list of downsamples to create. Should be strictly
            inceasing for maximum compatibility.
            Defaults to None.
        overwrite (bool, optional):
            Overwrite existing file. Defaults to False.
        verbose (bool, optional):
            Print more output. Defaults to False.
        ome (bool):
            Write OME-TIFF metadata. Defaults to False.
    """

    def __init__(
        self,
        path: Path,
        shape: Tuple[int, int],
        tile_size: Tuple[int, int] = (256, 256),
        dtype: np.dtype = np.uint8,  # Currently unused
        color_space: Optional[ColorSpace] = ColorSpace.RGB,
        codec: Optional[Codec] = Codec.JPEG,
        compression_level: int = 0,  # Currently unused
        microns_per_pixel: Tuple[float, float] = None,
        pyramid_downsamples: Optional[List[int]] = None,
        overwrite: bool = False,
        verbose: bool = False,
        **kwargs,
    ) -> None:
        # Validate inputs
        if dtype is not np.uint8:
            raise ValueError(f"SVSWriter only supports uint8, not {dtype}")
        if color_space not in (ColorSpace.RGB, ColorSpace.YCBCR):
            raise ValueError(
                f"SVSWriter only supports RGB and YCbCr, not {color_space}"
            )
        codec = Codec.from_string(codec)
        if codec not in (Codec.JPEG,):  # aperio_jp2000_ycbc not working
            raise ValueError(
                "SVSWriter currently only supports JPEG compession," f" not {codec}"
            )
        # Super
        super().__init__(
            path=path,
            shape=shape,
            tile_size=tile_size,
            dtype=dtype,
            color_space=color_space,
            codec=codec,
            compression_level=compression_level,
            microns_per_pixel=microns_per_pixel,
            pyramid_downsamples=pyramid_downsamples,
            overwrite=overwrite,
            verbose=verbose,
        )

    def __setitem__(self, index: Tuple[int, ...], value: np.ndarray) -> None:
        """Write pixel data at index. Not supported for TIFFWriter.

        In theory this is possible but it can be complex. If the new tile
        is larger in bytes, the tile will have to be added to the end of the
        file. The old tile will remain in the file and waste space unless it
        is later overwritten by another of length smaller or equal to the
        original tile.
        """
        raise NotImplementedError(
            "Compressed tiled TIFF files do not support random access writes."
        )

    def copy_from_reader(
        self,
        reader: Reader,
        num_workers: int = 2,
        read_tile_size: Optional[Tuple[int, int]] = None,
        timeout: float = 10.0,
    ) -> None:
        """Write pixel data to by copying from a Reader.

        Args:
            reader (Reader):
                Reader object.
            num_workers (int, optional):
                Number of workers to use. Defaults to 2.
            read_tile_size (Tuple[int, int], optional):
                Tile size to read. Defaults to None.
                This will use the tile size of the writer if None.
            timeout (float, optional):
                Timeout for workers. Defaults to 10s.
        """
        super().copy_from_reader(
            reader=reader,
            num_workers=num_workers,
            read_tile_size=read_tile_size,
            timeout=timeout,
        )
        import tifffile

        microns_per_pixel = self.microns_per_pixel or reader.microns_per_pixel
        resolution = (
            (
                round(mpp2ppu(microns_per_pixel[0], "cm")),
                round(mpp2ppu(microns_per_pixel[1], "cm")),
                "CENTIMETER",
            )
            if microns_per_pixel
            else None
        )

        with ZarrIntermediate(
            None, reader.shape, zero_after_read=False
        ) as intermediate:
            reader_tile_iterator = self.reader_tile_iterator(
                reader=reader,
                num_workers=num_workers,
                intermediate=intermediate,
                read_tile_size=read_tile_size or self.tile_size,
                timeout=timeout,
            )
            reader_tile_iterator = self.level_progress(reader_tile_iterator)
            # Write baseline (level 0)
            with tifffile.TiffWriter(
                self.path,
                bigtiff=True,
            ) as tif:

                # Construct the pipe separated Aperio description
                # Example description:
                # skipcq: PYL-W0105
                """
                Aperio Image Library v11.2.1\n
                46000x32914 [42673,5576 2220x2967] (240x240) JPEG/RGB Q=30;
                Aperio Image Library v10.0.51\n
                46920x33014 [0,100 46000x32914] (256x256) JPEG/RGB Q=30|
                AppMag = 20|
                StripeWidth = 2040|
                ScanScope ID = CPAPERIOCS|
                Filename = CMU-1|
                Date = 12/29/09|Time = 09:59:15|
                User = b414003d-95c6-48b0-9369-8010ed517ba7|
                Parmset = USM Filter|
                MPP = 0.4990|
                Left = 25.691574|Top = 23.449873|
                LineCameraSkew = -0.000424|
                LineAreaXOffset = 0.019265|LineAreaYOffset = -0.000313|
                Focus Offset = 0.000000|
                ImageID = 1004486|
                OriginalWidth = 46920|Originalheight = 33014|
                Filtered = 5|
                OriginalWidth = 46000|
                OriginalHeight = 32914
                """
                aperio_desc_compression = {
                    Codec.JPEG: f"JPEG/{self.color_space.upper()}",
                    Codec.JPEG2000: "J2K/YUV16",
                    "APERIO_JP2000_YCBC": "J2K/YUV16",
                }
                software = f"Aperio wsic Library v{wsic_version}"
                # Using reader shape for now, could be user specified in future
                original_height = reader.shape[0]
                original_width = reader.shape[1]
                # Not sure what these are copying original for now
                mystery_height = original_height
                mystery_width = original_width
                # Get compression in Aperio header format
                aperio_header_compression = aperio_desc_compression[self.codec.upper()]
                headers = [
                    (
                        f"{software} \n"
                        f"{original_width}x{original_height} "
                        f"[0,100 {mystery_width}x{mystery_height}] "
                        f"({self.tile_size[0]}x{self.tile_size[1]})"
                        f" {aperio_header_compression} "
                        f"Q={self.compression_level}"
                    )
                ]
                key_values = {
                    "AppMag": 20,  # e.g. 20
                    "StripeWidth": None,
                    "ScanScopeID": None,  # e.g. CPAPERIOCS
                    "Filename": None,
                    "Date": None,  # e.g. 01/01/22
                    "Time": None,  # e.g. 09:00:00
                    "User": None,  # e.g. UUID4
                    "Parmset": None,  # e.g. USM Filter
                    "MPP": np.mean(self.microns_per_pixel)
                    if self.microns_per_pixel
                    else None,
                    "Left": None,
                    "Top": None,
                    "LineCameraSkew": None,  # e.g. -0.003035
                    "LineAreaXOffset": None,  # e.g. 0.000000
                    "LineAreaYOffset": None,  # e.g. 0.000000
                    "Focus Offset": None,  # e.g. -0.001000
                    "DSR ID": None,  # e.g. homer
                    "ImageID": None,  # e.g. 1234
                    "OriginalWidth": original_width,
                    "OriginalHeight": original_height,
                    "Filtered": None,  # e.g. 5
                }
                description = (
                    "\n".join(headers)
                    + ("|" if any(key_values.values()) else "")
                    + "|".join(
                        f"{key} = {value}" for key, value in key_values.items() if value
                    )
                )

                compression = self.codec.condensed()
                tif.write(
                    data=iter(reader_tile_iterator),
                    tile=self.tile_size,
                    shape=reader.shape,
                    dtype=reader.dtype,
                    photometric=self.color_space,
                    compression=(compression, self.compression_level),
                    resolution=resolution,
                    subifds=len(self.pyramid_downsamples),
                    description=description,
                    subfiletype=0,
                )

                # Write thumbnail
                # NOTE: Assuming YXC order
                thumb_scale = max(
                    scale_to_fit(reader.shape[:2], (1024, 768)),
                    scale_to_fit(reader.shape[:2], (768, 1024)),
                )
                thumb_shape = tuple(floor(s * thumb_scale) for s in reader.shape[:2])
                thumbnail = reader.thumbnail(thumb_shape, approx_ok=True)
                tif.write(
                    data=thumbnail,
                    rowsperstrip=16,
                    dtype=np.uint8,
                    photometric=ColorSpace.RGB,
                    compression=Codec.JPEG,
                    description=software,
                    subfiletype=0,
                )

                # Write pyramid resolutions
                with multiprocessing.Pool(num_workers) as pool:
                    for level, downsample in self.pyramid_progress(
                        enumerate(self.pyramid_downsamples),
                        total=len(self.pyramid_downsamples),
                    ):
                        level_shape = tuple(
                            floor(s / downsample) for s in reader.shape[:2]
                        ) + (reader.shape[-1],)

                        level_tiles_shape = mosaic_shape(
                            level_shape,
                            self.tile_size,
                        )

                        func = partial(
                            get_level_tile,
                            tile_size=self.tile_size,
                            downsample=downsample,
                            read_intermediate_path=intermediate.path,
                        )

                        tile_generator = pool.imap(
                            func=func,
                            iterable=np.ndindex(level_tiles_shape),
                        )

                        tile_generator = self.level_progress(
                            tile_generator,
                            total=int(np.product(level_tiles_shape)),
                            desc=f"Level {level}",
                            leave=False,
                        )

                        tif.write(
                            data=iter(tile_generator),
                            tile=self.tile_size,
                            shape=level_shape,
                            dtype=reader.dtype,
                            photometric=self.color_space,
                            compression=(
                                compression,
                                self.compression_level,
                            ),
                            description=software,  # Optional for OpenSlide
                            subfiletype=1,  # Subfile type: 1 = reduced resolution
                        )


class ZarrReaderWriter(Writer, Reader):
    """Zarr reader and writer.

    Args:
        path (PathLike):
            Path to the output zarr.
        shape (Tuple[int, int]):
            Shape of the output zarr.
        tile_size (Tuple[int, int], optional):
            A (width, height) tuple of zarr chunks in pixels.
            Defaults to (256, 256).
        dtype (np.dtype, optional):
            Data type of the output zarr. Defaults to np.uint8.
        codec (str, optional):
            Compression codec to use. Defaults to None. Not all
            writers support compression.
        color_space (ColorSpace, optional):
            Color space. Defaults to "rgb".
        compression_level (int, optional):
            Compression level to use. Defaults to 0 (lossless /
            maximum).
        microns_per_pixel (Tuple[float, float], optional):
            A (width, height) tuple of microns per pixel. Defaults to
            None.
        pyramid_downsamples (List[int], optional):
            A list of downsamples to use in the pyramid. Defaults to
            None.
        overwrite (bool, optional):
            Overwrite output file if it exists.
            Defaults to False.
        verbose (bool, optional):
            Print more output. Defaults to False.
        ome (bool):
            Write OME-NGFF metadata. Defaults to False.
            Currently not implemented.

    """

    def __init__(
        self,
        path: Path,
        shape: Optional[Tuple[int, int]] = None,
        tile_size: Tuple[int, int] = (256, 256),
        dtype: np.dtype = np.uint8,
        color_space: Optional[ColorSpace] = "rgb",  # Currently unused
        codec: Union[str, Codec] = Codec.BLOSC,
        compression_level: int = 9,
        microns_per_pixel: Tuple[float, float] = None,  # Currently unused
        pyramid_downsamples: Optional[List[int]] = None,  # Currently unused
        overwrite: bool = False,
        verbose: bool = False,
        *,
        ome: bool = False,
    ) -> None:
        if color_space != "rgb":
            warn_unused(color_space)
        warn_unused(microns_per_pixel)
        super().__init__(
            path=path,
            shape=shape,
            tile_size=tile_size,
            dtype=dtype,
            color_space=color_space,
            codec=codec,
            compression_level=compression_level,
            microns_per_pixel=microns_per_pixel,
            pyramid_downsamples=pyramid_downsamples,
            overwrite=True,
            verbose=verbose,
        )
        self.ome = ome
        self.overwrite = overwrite
        register_codecs()
        self.compressor = self.get_codec(codec, compression_level)
        if self.path.exists() and not self.path.is_dir():
            raise FileExistsError(
                f"{self.path} exists but is not a directory. Zarrs must be directories."
            )

        self.zarr = None
        self._init_zarr(create=False)

    @property
    def tile_shape(self) -> Optional[Tuple[int, int]]:
        return self.zarr[0].chunks[:2] if self.zarr else None

    @property
    def mosaic_shape(self) -> Optional[Tuple[int, int]]:
        return mosaic_shape(self.shape, self.tile_shape)

    def _init_zarr(self, create: bool = True) -> zarr.Group:
        """Initialize the zarr.

        If the zarr already exists, it will be opened. Otherwise, it will be
        created if there is a shape.

        Args:
            create (bool):
                Create the zarr if it does not exist.

        Returns:
            zarr.Group:
                The zarr.
        """
        # Read an existing zarr
        if self.path.is_dir():
            self.zarr = zarr.open(
                self.path,
                mode="r+" if self.overwrite else "r",
            )
        # Create a new zarr group with one array if a shape was given
        if create:
            self.zarr = zarr.open_group(
                zarr.NestedDirectoryStore(self.path),
                mode="a",
            )
            self.zarr[0] = zarr.zeros(
                shape=self.shape,
                chunks=self.tile_size,
                dtype=self.dtype,
                compressor=self.compressor,
            )
        # If it is an array zarr, put it in a group
        if isinstance(self.zarr, zarr.Array):
            group = zarr.group()
            group[0] = self.zarr
            self.zarr = group
        # Get the shape and dtype from the zarr if possible
        if self.zarr is not None:
            self.shape = self.zarr[0].shape
            self.dtype = self.zarr[0].dtype
        # self.zarr may be None if the zarr was not created (no shape given)
        return self.zarr

    def get_codec(
        self,
        codec: Union[str, Codec],
        compression_level: int,
    ) -> Callable[[bytes], bytes]:
        """Get a codec for the given compression method and compression level."""
        from numcodecs import LZ4, LZMA, Blosc, Zlib, Zstd

        codec = Codec.from_string(codec)

        numcodecs_codecs = {
            Codec.LZ4: LZ4,
            Codec.LZMA: LZMA,
            Codec.BLOSC: Blosc,  # zarr default (lz4)
            Codec.ZLIB: Zlib,
            Codec.ZSTD: Zstd,
        }

        try:
            import imagecodecs

            imagecodecs_codecs = {
                Codec.DEFLATE: imagecodecs.numcodecs.Deflate,
                Codec.WEBP: imagecodecs.numcodecs.Webp,
                Codec.JPEG: imagecodecs.numcodecs.Jpeg,
                Codec.JPEGLS: imagecodecs.numcodecs.JpegLs,
                Codec.JPEG2000: imagecodecs.numcodecs.Jpeg2k,
                Codec.JPEGXL: imagecodecs.numcodecs.JpegXl,
                Codec.PNG: imagecodecs.numcodecs.Png,
                Codec.ZFP: imagecodecs.numcodecs.Zfp,
            }
        except ImportError:
            if self.verbose:
                print("imagecodecs not installed")
            imagecodecs_codecs = {}

        if codec in numcodecs_codecs:
            return numcodecs_codecs[codec](clevel=compression_level)

        if codec in imagecodecs_codecs:
            return imagecodecs_codecs[codec](level=compression_level)

        if codec == "qoi":
            from wsic.codecs import QOI

            return QOI()

        raise ValueError(f"Compression {codec} not supported.")

    def __setitem__(self, index: Tuple[int, ...], value: np.ndarray) -> None:
        """Write pixel data at index."""
        self.zarr[0][index] = value

    def __getitem__(self, index: Tuple[int, ...]) -> np.ndarray:
        """Read pixel data at index."""
        return self.zarr[0][index]

    def copy_from_reader(
        self,
        reader: Reader,
        num_workers: int = 2,
        read_tile_size: Optional[Tuple[int, int]] = None,
        timeout: float = 10.0,
        downsample_method: Optional[str] = None,
    ) -> None:
        """Write pixel data to by copying from a Reader.

        Args:
            reader (Reader):
                Reader object.
            num_workers (int, optional):
                Number of workers to use. Defaults to 2.
            read_tile_size (Tuple[int, int], optional):
                Tile size to read. Defaults to None.
                This will use the tile size of the writer if None.
            timeout (float, optional):
                Timeout for workers. Defaults to 10s.
            downsample_method (str, optional):
                Downsample method to use. Defaults to None.
                Valid downsample methods are: "cv2", "scipy", "np", None.
        """
        super().copy_from_reader(
            reader=reader,
            num_workers=num_workers,
            read_tile_size=read_tile_size,
            timeout=timeout,
        )

        # Ensure there is a zarr to write to
        if self.shape is None:
            self.shape = reader.shape
        self._init_zarr()

        # Validate and normalise inputs
        lossy_codecs = ["jpeg"]
        optionally_lossy_codecs = ["jpeg2000", "webp", "jpegls", "jpegxl", "jpegxr"]
        lossy = self.codec.condensed().lower() in lossy_codecs or (
            self.codec in optionally_lossy_codecs and self.compression_level > 0
        )
        read_tile_size = read_tile_size or self.tile_size
        write_multiple_of_read = all(np.mod(read_tile_size, self.tile_size) == 0)
        if lossy and not write_multiple_of_read:
            raise ValueError(
                "Lossy compression requires that the tile write size is a "
                "multiple of the read tile size."
            )

        # Create a reader tile iterator
        reader_tile_iterator = self.reader_tile_iterator(
            reader,
            read_tile_size=read_tile_size,
            yield_tile_size=read_tile_size,
            num_workers=num_workers,
            timeout=timeout,
        )
        reader_tile_iterator = self.level_progress(reader_tile_iterator)

        # Write the reader tile iterator to the writer
        tiles_shape = mosaic_shape(
            reader.shape,
            read_tile_size[::-1],
        )
        tiles_index = np.ndindex(tiles_shape)
        for ji, tile in zip(tiles_index, reader_tile_iterator):
            level_0 = self.zarr[0]
            level_0[tile_slices(ji, read_tile_size)] = tile

        self._build_pyramid(downsample_method)
        self._write_ome_metadata()

    def _write_ome_metadata(self) -> None:
        """Write OME-NGFF metadata to the .zattrs file in the root.

        This is based on version 0.4: https://ngff.openmicroscopy.org/0.4/.
        """
        if self.ome:
            multiscales = [
                ngff.Multiscales(
                    datasets=[
                        ngff.Dataset(
                            path=str(level),
                            coordinateTransformations=[
                                ngff.CoordinateTransform(
                                    "scale",
                                    [
                                        1,
                                        self.microns_per_pixel[0] * downsample,
                                        self.microns_per_pixel[1] * downsample,
                                    ],
                                )
                            ],
                        )
                        for level, downsample in enumerate(
                            [1] + self.pyramid_downsamples
                        )
                    ]
                    if self.microns_per_pixel is not None
                    else [],
                    axes=[
                        ngff.Axis("y", "space", "micronmeter"),
                        ngff.Axis("x", "space", "micronmeter"),
                        ngff.Axis("c", "channel", None),
                    ],
                )
            ]
            # Convert dataclasses
            meta_dict = dataclasses.asdict(
                ngff.Zattrs(multiscales=multiscales),
                # Exclude None values
                dict_factory=lambda x: {k: v for (k, v) in x if v is not None},
            )
            # Set the attrs
            for key, value in meta_dict.items():
                self.zarr.attrs[key] = value

    def _build_pyramid(self, downsample_method: Optional[str] = None):
        """Build the pyramid.

        Constructs additional levels of the pyramid from the first level.

        Args:
            downsample_method (str, optional):
                Downsample method to use. Defaults to None.

        """
        previous_level = self.zarr[0]
        previous_downsample = 1
        for level, downsample in self.pyramid_progress(
            enumerate(self.pyramid_downsamples, start=1),
        ):
            inter_level_downsample = downsample // previous_downsample
            # NOTE: Assuming length three shape with channels last
            level_shape = downsample_shape(self.shape, (downsample, downsample, 1))
            level_tiles_shape = mosaic_shape(
                level_shape,
                self.tile_size,
            )
            level_array = self.zarr.zeros(
                name=level,
                shape=level_shape,
                chunks=(*self.tile_size, self.shape[-1]),
                dtype=self.dtype,
                compressor=self.compressor,
            )
            level_tiles_index = np.ndindex(level_tiles_shape)

            level_read_tile_size = np.multiply(self.tile_size, inter_level_downsample)

            # Write tiles to the level by copying from the previous level
            for ji in self.level_progress(level_tiles_index):
                read_slices = tile_slices(ji, level_read_tile_size)
                tile = previous_level[read_slices]
                down_tile = downsample_tile(
                    tile, inter_level_downsample, method=downsample_method
                )
                write_slices = tile_slices(ji, self.tile_size)
                level_array[write_slices] = down_tile
            previous_level = level_array
            previous_downsample = downsample

    def transcode_from_reader(
        self,
        reader: Union[TIFFReader, DICOMWSIReader],
        downsample_method: Optional[str] = None,
    ) -> None:
        """Losslessly transform into a new format from a supported Reader.

        Repackages tiles from the Reader to a zarr. Currently only
        supports transcoding from:

        - JPEG compressed SVS (:class:`wsic.readers.TIFFReader`)
        - J2K compressed SVS (:class:`wsic.readers.TIFFReader`)
        - JPEG compressed OME-TIFF (:class:`wsic.readers.TIFFReader`)
        - JPEG compressed DICOM WSI (:class:`wsic.readers.DICOMWSIReader`)
        - JPEG compressed NDPI (Hamamatsu)

        Currently only outputs a single resolution level (level 0).

        It may also be possible to transcode the tiles themselves (e.g.
        JPEG to JPEG XL) or perform simple geometric transforms (flip,
        rotate, etc). However, this is not yet implemented. Currently,
        they are simply copied into a new structure.


        Args:
            reader (Reader):
                Reader object.
            downsample_method (str, optional):
                Downsample method to use for new reduced resolutions.
                Defaults to None.
                Valid downsample methods are: "cv2", "scipy", "np", None.
        """
        transcode_supported = self._can_transcode_from_reader(reader)
        if not transcode_supported:
            raise ValueError(
                "Currently only SVS, NDPI, OME-TIFF, and WSI DICOM "
                "(with JPEG, JPEG2000, or WebP compression) "
                "are supported for transcoding."
            )

        register_codecs()
        codec = self.get_transcode_codec(reader)
        self.shape = reader.shape
        self.dtype = reader.dtype

        self.zarr = zarr.open_group(zarr.NestedDirectoryStore(self.path))
        self.zarr.create_dataset(
            name="0",
            shape=self.shape,
            dtype=self.dtype,
            chunks=(*reader.tile_shape, reader.shape[-1]),
            compressor=codec,
        )

        # Copy tiles
        for index in self.transcode_progress(
            np.ndindex(reader.mosaic_shape),
            total=np.prod(reader.mosaic_shape),
        ):
            tile_path = self.path / "0" / str(index[0]) / str(index[1]) / "0"
            tile_path.parent.mkdir(parents=True, exist_ok=True)
            tile_bytes = reader.get_tile(index, decode=False)
            with open(tile_path, "wb") as file_handle:
                file_handle.write(tile_bytes)

        self._build_pyramid(downsample_method)
        self._write_ome_metadata()

    def _can_transcode_from_reader(self, reader: Reader) -> bool:
        """Determine if a reader supports from to the current writer.

        Args:
            reader (Reader):
                Reader object.

        Returns:
            bool:
                Whether the reader supports being transcoded from.
        """
        # 1. A valid get_tile(decode=False)
        try:
            reader.get_tile((0, 0), decode=False)
        except (NotImplementedError, AttributeError):
            raise ValueError(
                "Reader must have a get_tile method which can return encoded tiles"
                " (decoded=False)."
            )
        # 2. Compatible tile sizes
        if self.tile_size != reader.tile_shape[:2][::-1]:
            raise ValueError(
                "Tile size must match the reader tile size for transcoding."
            )
        # 3. Matching data types
        if self.dtype != reader.dtype:
            raise ValueError("Dtype must match the reader dtype for transcoding.")
        # 4. Compatible compression
        has_valid_compression = (
            Codec.from_string(reader.codec) in (Codec.JPEG, Codec.JPEG2000, Codec.WEBP),
        )
        # 5. Known supported TIFF formats
        is_generic_tiff = isinstance(reader, TIFFReader) and (
            reader.tiff.pages[0].is_tiled and has_valid_compression
        )
        is_supported_tiff = isinstance(reader, TIFFReader) and any(
            [
                reader.tiff.is_svs,
                reader.tiff.is_ome,
                reader.tiff.is_ndpi,
                is_generic_tiff,
            ]
        )
        # 5. Supported Reader (WSIDICOM or a TIFF with supported format)
        return all(
            [
                isinstance(reader, (TIFFReader, DICOMWSIReader)),
                has_valid_compression,
                not isinstance(reader, TIFFReader) or is_supported_tiff,
            ]
        )

    @staticmethod
    def get_transcode_codec(reader: TIFFReader) -> Any:
        """Get the codec to use for transcoding.

        Args:
            reader (TiffReader):
                Reader object.

        Returns:
            numcodecs.Codec:
                Codec to use for transcoding.
        """
        from imagecodecs.numcodecs import Jpeg, Jpeg2k, Webp

        # Try to get the compression level from the reader if known
        try:
            level = reader.compression_level
        except AttributeError:
            level = None

        # Create the codec object
        if reader.codec == Codec.JPEG:
            return Jpeg(
                tables=reader.jpeg_tables,
                colorspace_jpeg=reader.color_space,
                colorspace_data=ColorSpace.RGB,
                level=reader.compression_level,
            )
        if reader.codec == Codec.JPEG2000:
            return Jpeg2k(codecformat="JP2", colorspace=reader.color_space, level=level)
        if reader.codec == Codec.WEBP:
            return Webp(level=level)
        # Out of options
        raise ValueError(
            "Currently only JPEG, J2K (JPEG-2000), and WebP compression "
            " are supported for transcoding."
        )


class ZarrIntermediate(Writer, Reader):
    """Zarr intermediate reader/writer.

    A convenience reader/writer which is also a context manager. This
    allows for changing of tile order or size when converting between
    formats and also avoids decoding the same tile from the original
    file twice. This is particularly useful for formats which are very
    computationally costly to decode such as JPEG 2000.

    Args:
        path (PathLike):
            Path to the intermediate file. If None, a temporary file
            will be created.
        shape (Tuple[int, int]):
            Shape of the output file.
        tile_size (Tuple[int, int], optional):
            A (width, height) tuple of zarr chunk size in pixels.
            Defaults to (256, 256).
        dtype (np.dtype, optional):
            The data type of the output file. Defaults to np.uint8.
        color_space (ColorSpace, optional):
            Unused but kept for compatibility with the Writer base
            class.
        compression (str, optional):
            Unused but kept for compatibility with the Writer base
            class. Internally uses default zarr compression.
        compression_level (int, optional):
            Unused but kept for compatibility with the  Writer base
            classes. Internally uses default zarr compression level.
        microns_per_pixel (float, optional):
            Unused but kept for compatibility with the Reader and Writer
            classes.
        pyramid_downsamples (List[int], optional):
            Unused but kept for compatibility with the Reader and Writer
            classes.
        overwrite (bool, optional):
            If True, the output file will be overwritten if it exists.
            Defaults to False.
        verbose (bool, optional):
            If True, print information about the file being written.
        zero_after_write (bool, optional):
            If True, data in the zarr will be zeroed after writing.
            Defaults to False.
    """

    def __init__(
        self,
        path: PathLike,
        shape: Tuple[int, int],
        tile_size: Tuple[int, int] = (256, 256),
        dtype: np.dtype = np.uint8,
        color_space: Optional[ColorSpace] = "rgb",  # Currently unused
        codec: Optional[Codec] = None,  # Currently unused
        compression_level: int = 0,  # Currently unused
        microns_per_pixel: Tuple[float, float] = None,  # Currently unused
        pyramid_downsamples: Optional[List[int]] = None,  # Currently unused
        overwrite: bool = False,
        verbose: bool = False,
        *,
        zero_after_read: bool = False,
    ) -> None:
        if color_space != "rgb":
            warn_unused(color_space)
        warn_unused(codec)
        warn_unused(compression_level, ignore_falsey=True)
        warn_unused(microns_per_pixel)
        warn_unused(pyramid_downsamples, ignore_falsey=True)
        # Create a temporary path if no path is given
        path = path or Path(tempfile.gettempdir(), uuid.uuid4().hex).with_suffix(
            ".zarr"
        )
        super().__init__(
            path=path,
            shape=shape,
            tile_size=tile_size,
            dtype=dtype,
            color_space=color_space,
            codec=codec,
            compression_level=compression_level,
            microns_per_pixel=microns_per_pixel,
            pyramid_downsamples=pyramid_downsamples,
            overwrite=overwrite,
            verbose=verbose,
        )
        self.zero_after_read = zero_after_read

        self.path.mkdir(parents=True, exist_ok=True)

        self.zarr = zarr.open(
            store=zarr.NestedDirectoryStore(path),
            mode="a",
            shape=self.shape,
            chunks=(*self.tile_size, self.shape[-1]),
            dtype=self.dtype,
        )

    def __setitem__(
        self, index: Tuple[Union[int, slice], ...], value: np.ndarray
    ) -> None:
        """Write pixel data at index."""
        self.zarr[index] = value

    def __getitem__(self, index: Tuple[int, ...]) -> np.ndarray:
        """Read pixel data at index."""
        result = self.zarr[index]
        if self.zero_after_read:
            self.zarr[index] = 0
        return result  # noqa: R504

    def __enter__(self) -> "ZarrIntermediate":
        """Enter the context."""
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """Exit the context."""
        if self.path.exists():
            shutil.rmtree(self.path)

    def copy_from_reader(
        self,
        reader: Reader,
        num_workers: int = 2,
        read_tile_size: Optional[Tuple[int, int]] = None,
        timeout: float = 10.0,
    ) -> None:
        """Not supported but included for API consistency."""
        raise NotImplementedError()


def _cv2_downsample(image: np.ndarray, factor: int) -> np.ndarray:
    """Resample an image using OpenCV.

    Args:
        image (np.ndarray):
            The image to resample.
        factor (int):
            The downsampling factor.

    Returns:
        np.ndarray:
            The resampled image.
    """
    import cv2

    return cv2.resize(
        image,
        (image.shape[1] // factor, image.shape[0] // factor),
        interpolation=cv2.INTER_AREA,
    )


def _pil_downsample(image: np.ndarray, factor: int) -> np.ndarray:
    """Resample an image using PIL.

    Args:
        image (np.ndarray):
            The image to resample.
        factor (int):
            The downsampling factor.

    Returns:
        np.ndarray:
            The resampled image.
    """
    import PIL.Image

    return PIL.Image.fromarray(image).resize(
        (image.shape[1] // factor, image.shape[0] // factor),
        resample=PIL.Image.Resampling.Box,
    )


def _scipy_downsample(image: np.ndarray, factor: int) -> np.ndarray:
    """Resample an image using SciPy.

    Args:
        image (np.ndarray):
            The image to resample.
        factor (int):
            The downsampling factor.

    Returns:
        np.ndarray:
            The resampled image.
    """
    from scipy import ndimage

    return ndimage.zoom(image, (1 / factor, 1 / factor, 1), order=1)


def _np_downsample(image: np.ndarray, factor: int) -> np.ndarray:
    """Resample an image using NumPy.

    Args:
        image (np.ndarray):
            The image to resample.
        factor (int):
            The downsampling factor.

    Returns:
        np.ndarray:
            The resampled image.
    """
    return mean_pool(image.astype(np.float), factor).clip(0, 255).astype(np.uint8)


def downsample_tile(
    image: np.ndarray, factor: int, method: Optional[str] = None
) -> np.array:
    """Downsample an image by a factor.

    Args:
        image (np.ndarray):
            The image to downsample.
        factor (int):
            The downsampling factor.
        method (str):
            The downsampling method (library) to use.
            Defaults to None, which tries cv2, then scipy, and falls
            back to numpy.
            Valid options are: "cv2", "pillow", "scipy", "np", None.
    """
    methods = {
        "cv2": _cv2_downsample,
        "pillow": _pil_downsample,
        "scipy": _scipy_downsample,
        "np": _np_downsample,
    }

    if method is not None and method not in methods:
        raise ValueError(f"Invalid method: {method}")

    if method in methods:
        return methods[method](image, factor)

    for method_name, func in methods.items():
        try:
            return func(image, factor)
        except ImportError:
            warnings.warn(
                f"Failed to import library for {method_name} for downsampling. "
                "It may not be installed."
            )
    raise Exception("Failed to use any downsampling method.")


def get_level_tile(
    yx: Tuple[int, int],
    tile_size: Tuple[int, int],
    downsample: int,
    read_intermediate_path: PathLike,
    downsample_method: Optional[str] = None,
) -> np.ndarray:
    """Generate tiles for a downsampled level.

    Args:
        yx (Tuple[int, int]):
            The tile coordinates.
        tile_size (Tuple[int, int]):
            The tile size.
        downsample (int):
            The downsampling factor.
        read_intermediate_path (PathLike):
            The path to the intermediate file (zarr).
        downsample_method (str):
            The downsampling method (library) to use.
    """
    y, x = yx
    w, h = tile_size
    tile_index = (
        slice(y * h * downsample, (y + 1) * h * downsample),
        slice(x * w * downsample, (x + 1) * w * downsample),
    )
    reader = zarr.open(read_intermediate_path, mode="r")
    tile = reader[tile_index]
    return downsample_tile(tile, downsample, method=downsample_method)



"""Enumerated types used by wsic."""
from enum import Enum
from typing import Optional


class Codec(str, Enum):
    """Compression codecs / algorithms / formats."""

    AVIF = "AVIF"  # AV1 Image Format (HEIF)
    BLOSC = "Blosc"  # Block based meta-compression algorithm
    BLOSC2 = "Blosc2"  # Block based meta-compression version 2
    BROTLIT = "Brotli"  # Google Brotli
    BZ2 = "BZ2"  # Bzip2
    DEFLATE = "DEFLATE"
    DELTA = "Delta"  # Delta coding
    GIF = "GIF"  # Graphics Interchange Format
    GZIP = "GZIP"  # Gzip
    J2K = "J2K"  # JPEG 2000 raw codestream
    JPEG = "JPEG"  # Original JPEG
    JPEG2000 = "JPEG 2000"
    JPEGLS = "JPEG-LS"  # Lossless (or near) JPEG
    JPEGXL = "JPEG XL"
    JPEGXR = "JPEG XR"
    LERC = "LERC"  # Limited Error Raster Compression
    LJPEG = "LJPEG"  # Lossy JPEG (old version)
    LZ4 = "LZ4"
    LZ4F = "LZ4F"
    LZ77 = "LZ77"
    LZMA = "LZMA"  # Lempel-Ziv-Welch chain algorithm
    LZW = "LZW"  # Lempel-Ziv-Welch
    NONE = "None"  # No compression
    PACKBITS = "PackBits"
    PNG = "PNG"  # Portable Network Graphics
    QOI = "QOI"  # Quite OK Image
    SNAPPY = "Snappy"  # Google Snappy
    WEBP = "WebP"
    ZFP = "ZFP"  # Floating point compression (up to 4 dimensions)
    ZFPY = "ZFPY"  # Python version of ZFP
    ZLIB = "Zlib"
    ZLIBNG = "ZlibNG"  # ZlibNG, zlib replacement for "next generation" systems
    ZOPFLI = "Zopfli"
    ZSTD = "Zstandard"  # Zstandard
    ISO_10918_1 = "JPEG"  # noqa: PIE796
    ISO_15444_1 = "JPEG 2000"  # noqa: PIE796
    ISO_14495_1 = "JPEG-LS"  # noqa: PIE796

    def condensed(self) -> str:
        """Convert to a string without spaces or dashes."""
        return self.value.replace(" ", "").replace("-", "")

    def numcodecs_id(self) -> str:
        """Convert to numcodecs Codec ID string."""
        if self in WSIC_CODECS:
            return "imagecodecs_" + self.condensed().lower()
        if self in NUMCODECS_CODECS:
            return self.condensed.lower()
        if self in IMAGECODECS_CODECS:
            return "imagecodecs_" + self.condensed().lower()
        return self.condensed().lower()

    @classmethod
    def from_string(cls, string: str) -> "Codec":
        """Convert string to Compression enum."""
        condensed_upper = string.replace(" ", "").replace("-", "").upper()
        try:
            return getattr(cls, condensed_upper)
        except AttributeError:
            raise ValueError(f"Unknown compression: {string}")

    @classmethod
    def from_tiff(cls, compression: int) -> "Codec":
        """Convert TIFF compression value to Compression enum.

        Args:
            compression:
                TIFF compression value.
        """
        compression_codec_mapping = {
            1: cls.NONE,
            5: cls.LZW,
            7: cls.JPEG,
            34712: cls.JPEG2000,
            33003: cls.JPEG2000,  # Leica Aperio YCBC
            33005: cls.JPEG2000,  # Leica Aperio RGB
            34934: cls.JPEGXR,
            22610: cls.JPEGXR,
            34927: cls.WEBP,  # Deprecated
            50001: cls.WEBP,
            34926: cls.ZSTD,  # Deprecated
            50000: cls.ZSTD,
            50002: cls.JPEGXL,
        }

        if compression in compression_codec_mapping:
            return compression_codec_mapping[compression]
        raise ValueError(f"Unknown TIFF compression: {compression}")


WSIC_CODECS = (Codec.QOI,)

NUMCODECS_CODECS = (
    Codec.BLOSC,
    Codec.DEFLATE,
    Codec.DELTA,
    Codec.GZIP,
    Codec.LZ4,
    Codec.LZMA,
    Codec.PACKBITS,
    Codec.ZLIB,
    Codec.ZSTD,
    Codec.ZFPY,
)

IMAGECODECS_CODECS = (
    Codec.AVIF,
    Codec.BROTLIT,
    Codec.DEFLATE,
    Codec.DELTA,
    Codec.GIF,
    Codec.J2K,
    Codec.JPEG,
    Codec.JPEG2000,
    Codec.JPEGLS,
    Codec.JPEGXL,
    Codec.JPEGXR,
    Codec.LERC,
    Codec.LJPEG,
    Codec.LZ4,
    Codec.LZ4F,
    Codec.LZMA,
    Codec.PNG,
    Codec.SNAPPY,
    Codec.WEBP,
    Codec.ZFP,
    Codec.ZLIB,
    Codec.ZLIBNG,
    Codec.ZOPFLI,
)


class ColorSpace(str, Enum):
    """Color spaces."""

    RGB = "RGB"  # Standard Red Green Blue, assumed to be sRGB
    SRGB = "RGB"  # noqa: PIE796
    RGBA = "RGBA"  #
    LINEAR = "Linear"  # Generic linear color space
    GREY = "Grey"  # Generic greyscale
    GRAY = "Grey"  # noqa: PIE796
    GREYSCALE = "Grey"  # noqa: PIE796
    GRAYSCALE = "Grey"  # noqa: PIE796
    MINISBLACK = "Grey"  # noqa: PIE796
    MINISWHITE = "Min is White"
    CMYK = "CMYK"  # Cyan Magenta Yellow Black
    CMYKA = "CMYKA"  # Cyan Magenta Yellow Black Alpha
    YCBCR = "YCbCr"  # Assumed to be BT.601 / Rec. 601
    YCRCB = "YCrCb"
    BT601 = "YCbCr"  # noqa: PIE796
    REC601 = "YCbCr"  # noqa: PIE796
    YCOCG = "YCoCg"  # Reversible modified YUV used by JPEG 2000, defined in ITU-T H.273
    YCC = "YCoCg"  # noqa: PIE796
    YUV = "YUV"  # BT.1700
    BT1700 = "YUV"  # noqa: PIE796 BT.1700
    CIELAB = "CIE L*a*b*"  # CIE L*a*b*
    CIELUV = "CIE L*u*v*"  # CIE L*u*v*
    HSV = "HSV"  # Hue Saturation Value
    HSL = "HSL"  # Hue Saturation Lightness
    LMS = "LMS"  # Cone-cone LMS
    XYB = "XYB"  # Used by JPEG XL, derived from LMS
    RLAB = "RLAB"  # https://en.wikipedia.org/wiki/Color_appearance_model#RLAB
    LLAB = "LLAB"  # https://en.wikipedia.org/wiki/Color_appearance_model#LLAB
    OKLAB = "OKLab"  # https://en.wikipedia.org/wiki/Color_appearance_model#OKLab
    PALETTE = "Palette"  # Used by PNG, JPEG, and TIFF

    def condensed(self) -> str:
        """Convert to a string without spaces, dashes, and asterisks."""
        return self.value.replace(" ", "").replace("-", "").replace("*", "")

    @classmethod
    def from_string(cls, string: str) -> "Codec":
        """Convert string to ColorSpace enum."""
        condensed_upper = (
            string.replace(" ", "").replace("-", "").replace("*", "").upper()
        )
        try:
            return getattr(cls, condensed_upper)
        except AttributeError:
            raise ValueError(f"Unknown color space: {string}")

    @classmethod
    def from_tiff(
        cls, photometric: int, compression: Optional[int] = None
    ) -> "ColorSpace":
        """Convert TIFF value to ColorSpace enum.

        Args:
            photometric:
                TIFF photometric value.
            compression:
                TIFF compression value.
        """
        compression_colorspace_mapping = {
            33003: cls.YCBCR,
            33005: cls.RGB,
        }

        photometric_color_space_mapping = {
            1: cls.GRAY,
            2: cls.RGB,
            3: cls.PALETTE,
            5: cls.CMYK,
            6: cls.YCBCR,
            8: cls.CIELAB,
            34892: cls.LINEAR,
        }
        if compression in compression_colorspace_mapping:
            return compression_colorspace_mapping[compression]
        if photometric in photometric_color_space_mapping:
            return photometric_color_space_mapping[photometric]
        raise ValueError(f"Unsupported TIFF photometric interpretation: {photometric}")

    @classmethod
    def from_dicom(cls, photometric: str) -> "ColorSpace":
        """Convert DICOM value to ColorSpace enum.

        Args:
            photometric:
                DICOM photometric value.
        """
        photometric_color_space_mapping = {
            "MONOCHROME1": cls.MINISBLACK,
            "MONOCHROME2": cls.MINISWHITE,
            "RGB": cls.RGB,
            "PALETTE_COLOR": cls.PALETTE,
            "YBR_FULL": cls.YCRCB,
            "YBR_FULL_422": cls.YCRCB,
            "YBR_PARTIAL_420": cls.YCRCB,
        }
        if photometric in photometric_color_space_mapping:
            return photometric_color_space_mapping[photometric]
        raise ValueError(f"Unsupported DICOM photometric interpretation: {photometric}")



"""Custom codecs for wsic."""
import warnings
from typing import IO

import numpy as np
from numcodecs.abc import Codec
from numcodecs.compat import ensure_ndarray


class QOI(Codec):
    """Quite OK Image (QOI) Format.

    See: https://qoiformat.org/
    """

    codec_id = "wsic_qoi"

    @staticmethod
    def encode(buf: IO) -> bytes:
        """Encode QOI data."""
        import qoi

        return qoi.encode(ensure_ndarray(buf))

    @staticmethod
    def decode(buf: IO, out=None) -> np.ndarray:
        """Decode QOI data."""
        import qoi

        if out is not None:
            out = qoi.decode(ensure_ndarray(buf).tobytes())
            return out  # noqa: PIE781, R504
        return qoi.decode(buf)


def register_codecs() -> None:
    """Add additional codecs to the numcodecs registry.

    Codecs from imagecodecs include:
    - JPEG
    - PNG
    - Deflate
    - JPEG 2000
    - JPEG-LS
    - JPEG XR
    - JPEG XL
    - WebP
    - Zfp

    Additional codecs from wsic:
    - QOI
    """
    import numcodecs

    # Register wsci codecs
    try:
        import qoi  # noqa: F401

        numcodecs.register_codec(QOI)
    except ImportError:
        pass

    try:
        from imagecodecs.numcodecs import register_codecs as register_imagecodecs_codecs

        # Register imagecodecs codecs
        if "imagecodecs_jpeg" not in numcodecs.registry.codec_registry:
            register_imagecodecs_codecs()
    except ImportError:
        warnings.warn(
            "imagecodecs is not installed, some codecs will not be available."
        )



import inspect
import warnings
from contextlib import suppress
from math import ceil, floor
from numbers import Number
from typing import Any, Callable, Dict, Iterable, Optional, Tuple, Union

import numpy as np


def downsample_shape(
    baseline_shape: Tuple[int, ...],
    downsample: int,
    rounding_func: Callable[[Number], int] = floor,
) -> Tuple[int, ...]:
    r"""Calculate the shape of an array after downsampling by a factor.

    The shape is calculated by dividing the shape of the baseline array
    by the downsample factor. The output is rounded to the nearest
    integer using the provided rounding function. E.g. for a founding
    function of `floor`, the following opertion is performed
    :math:`\lfloor \frac{shape}{downsample} \rfloor`. If a channel
    dimension is specified, the dimension is left unchanged.

    Args:
        baseline_shape (Tuple[int, ...]):
            The shape of the array to downsample.
        downsample (int):
            The downsample factor.
        rounding_func (Callable[[int], int]):
            The rounding function to use. Defaults to floor. Any
            function which takes a single Number and returns an int such
            as `math.floor` or `math.ceil` can be used. Note that the
            behaviour of floor differs for negative numbers, e.g.
            floor(-1) = -2. The `int` function is sigificantly faster
            than floor.

    Returns:
        Tuple[int, ...]:
            The shape of the downsampled array.

    Examples:
        >>> dowmsample_shape((13, 13), 2)
        (6, 6)

        >>> dowmsample_shape((13, 13, 3), 2)
        (6, 6, 3)

        >>> dowmsample_shape((13, 13, 3), 2, channel_dim=2)
        (6, 6, 3)

        >>> downsample_shape((13, 13, 3), 2, -1, ceil)
        (7, 7, 3)

    """
    if isinstance(downsample, Number):
        downsample = [downsample] * len(baseline_shape)
    if len(downsample) != len(baseline_shape):
        raise ValueError(
            f"The number of downsample factors ({len(downsample)}) "
            f"does not match the number of dimensions ({len(baseline_shape)})."
        )
    return tuple(rounding_func(x / s) for x, s in zip(baseline_shape, downsample))


def block_downsample_shape(
    shape: Tuple[int, ...], downsample: float, block_shape: Tuple[int, ...]
) -> Tuple[int, ...]:
    """Calculate the shape of an array after downsampling in fixed size chunks."""
    shape = np.array(shape)
    block_shape = np.array(block_shape)
    mosaic_shape_int = shape // block_shape
    new_tile_shape = downsample_shape(block_shape, downsample)
    edges_shape = shape - (mosaic_shape_int * block_shape)
    edge_tile_shapes = downsample_shape(edges_shape, downsample)
    new_shape = (
        np.array(mosaic_shape_int) * np.array(new_tile_shape)
    ) + edge_tile_shapes
    return tuple(new_shape.astype(int)), new_tile_shape


def scale_to_fit(shape: Tuple[int, ...], max_shape: Tuple[int, ...]) -> float:
    """Find the scale factor to fit shape into a max shape.

    Given a shape and a max shape, find the scale factor to apply to the
    shape to fit it within the max shape while preserving the aspect
    ratio.

    Args:
        shape (Tuple[int, ...]):
            The shape to fit.
        max_shape (Tuple[int, ...]):
            The maximum shape to fit into.

    Returns:
        float:
            The scale factor to apply to the shape to fit it within the
    """
    shape = np.array(shape)
    max_shape = np.array(max_shape)
    return np.min(max_shape / shape)


def varnames(
    var: Any,
    f_backs: int = 1,
    squeeze: bool = True,
) -> Optional[Union[Tuple[str], str]]:
    """Get the name(s) of a variable.

    A bit of a hack, but works for most cases. Good for debugging and
    making logging messages more helpful. Works by inspecting the call
    stack and finding the name of the variable in the caller's frame by
    checking the object's ID. There may be multiple variable names with
    the same ID and hence a tuple of name strings is returned.

    Args:
        var (Any):
            The variable to get the name of.
        f_backs (int):
            The number of frames to go back in the call stack.
        squeeze (bool):
            If only one name is found in the call frame, return it as a
            string instead of a tuple of strings

    Returns:
        Optional[Union[Tuple[str], str]]:
            The name(s) of the variable.

    Examples:
        >>> foo = "bar"
        >>> varnames(foo)
        foo

        >>> foo = "bar"
        >>> baz = foo
        >>> varnames(foo)
        (foo, baz)

        >>> varnames("bar")  # Literals will return None
        None
    """
    # Get parent (caller) frame
    call_frame = inspect.currentframe()
    for _ in range(f_backs):
        call_frame = call_frame.f_back
    # Find the name of the variable in the parent frame
    var_names = tuple(
        var_name
        for var_name, var_val in reversed(call_frame.f_locals.items())
        if var_val is var
    )
    if not squeeze or len(var_names) > 1:
        return var_names
    if len(var_names) == 1:
        return var_names[0]
    return None


def warn_unused(
    var: Any,
    name: Optional[str] = None,
    ignore_none: bool = True,
    ignore_falsey: bool = False,
) -> None:
    """Warn the user if a variable has a non None or non falsey value.

    See
    https://docs.python.org/3/library/stdtypes.html#truth-value-testing
    for an explanation of what evaluates to true and false.

    Used when some kwargs are defined for API consistency and to satisfy
    the Liskov Substitution Principle (LSP).

    Args:
        var (Any):
            The variable to check.
        name (Optional[str]):
            The name of the variable. If None, the variable name will be
            obtained from the call frame.
        ignore_none (bool):
            If True, do not warn if the variable is None.
        ignore_falsey (bool):
            If True, do not warn if the variable is any falsey value.
    """
    name = name or str(varnames(var, 2))
    if ignore_none and var is None:
        return
    if ignore_falsey and not var:
        return
    if var is not None:
        warnings.warn(
            f"Argument '{name}' is currently unsued and is being ignored.",
            stacklevel=2,
        )


def mpp2ppu(mpp: float, units: Union[str, int]) -> float:
    """Convert microns per pixel (mpp) to pixels per unit.

    Args:
        mpp (float):
            The microns per pixel.
        units (Union[str, int]):
            The units to convert to. Valid units are: 'um', 'mm', 'cm',
            'inch', 2 (TIFF inches), and 3 (TIFF cm).

    """
    mpp_to_upp = {
        "um": 1,
        "mm": 1e3,
        "cm": 1e6,
        "inch": 25400,
        2: 25400,
        3: 1e6,
    }
    return (1 / mpp) * mpp_to_upp[units]


def ppu2mpp(ppu: float, units: Union[str, int]) -> float:
    """Convert pixels per unit to microns per pixel (mpp).

    Args:
        ppu (float):
            The pixels per unit.
        units (Union[str, int]):
            The units to convert from. Valid units are: 'um', 'mm',
            'cm', 'inch', 2 (TIFF inches), and 3 (TIFF cm).

    """
    if ppu == 0:
        return 0
    mpp_to_upp = {
        "um": 1,
        "mm": 1e3,
        "cm": 1e4,
        "inch": 25400,
        2: 25400,
        3: 1e4,
    }
    return (1 / ppu) * mpp_to_upp[units]


def mosaic_shape(
    array_shape: Tuple[int, ...], tile_shape: Tuple[int, ...]
) -> Tuple[int, ...]:
    """Calculate the shape of a grid of tiles which covers an array.

    Args:
        shape (Tuple[int, ...]):
            The shape of the array to cover.
        tile_shape (Tuple[int, ...]):
            The shape of the tiles.

    Returns:
        Tuple[int, ...]:
            The shape of the tiles which cover shape.

    Examples:
        >>> tile_shape((13, 13), (8, 8))
        (2, 2)

        >>> tile_shape((13, 13, 3), (8, 8))
        (2, 2)

        >>> tile_shape((13, 13, 3), (8, 8, 3))
        (2, 2, 1)

    """
    return tuple(ceil(x / y) for x, y in zip(array_shape, tile_shape))


def strictly_increasing(iterable: Iterable) -> bool:
    """Check if an iterable is strictly increasing."""
    return all(x < y for x, y in zip(iterable, iterable[1:]))


def tile_slices(
    index: Tuple[int, ...],
    shape: Tuple[int, ...],
) -> Tuple[slice, ...]:
    """Create a tuple of slices to read a tile region from an array.

    Args:
        location (Tuple[int, ...]):
            The index of the tile e.g. the (ith, jth) tile in a 2d grid.
        shape (Tuple[int, ...]):
            The shape of the tiles in the grid.

    Returns:
        Tuple[slice, ...]:
            The slices to read the tile region from an array-like.
    """
    return tuple(slice(loc * s, (loc + 1) * s) for loc, s in zip(index, shape))


def wrap_index(
    index: Tuple[int, ...],
    shape: Tuple[int, ...],
    reverse: bool = True,
) -> Tuple[Tuple[int, ...], int]:
    """Wrap an index to the shape of an array.

    Args:
        index (Tuple[int, ...]):
            The index to wrap.
        shape (Tuple[int, ...]):
            The shape of the array.
        reverse (bool):
            If True, wrap the index to the opposite end of the array.

    Returns:
        Tuple[Tuple[int, ...], int]:
            The wrapped index and any overflow.

    Examples:
        >>> wrap_index((0, 3), (3, 3))
        ((1, 0), 0)

        >>> wrap_index((1, 4), (3, 3))
        ((2, 1), 0)

        >>> wrap_index((3, 1), (3, 3), reverse=False)
        ((0, 2), 0)

    """
    if len(index) != len(shape):
        raise ValueError("Index and shape must have the same number of dimensions.")

    wrapped = list(index[::-1]) if reverse else list(index)
    overflow = 0
    shape = reversed(shape) if reverse else shape
    for i, s in enumerate(shape):
        wrapped[i] += overflow
        overflow = wrapped[i] // s
        wrapped[i] = wrapped[i] % s
    wrapped = reversed(wrapped) if reverse else wrapped
    return tuple(wrapped), overflow


def view_as_blocks(array: np.ndarray, block_shape: Tuple[int, ...]) -> np.ndarray:
    """View an array as a grid of non-overlapping blocks.

    The same method as in scikit-image and several other libraries,
    using the `numpy.lib.stride_tricks.as_strided` function to produce a
    view.

    Args:
        array (np.ndarray):
            The array to view.
        block_shape (Tuple[int, ...]):
            The shape of the blocks.

    Returns:
        np.ndarray:
            The array view as a grid of non-overlapping blocks.
    """
    from numpy.lib.stride_tricks import as_strided

    block_shape = np.array(block_shape)
    new_shape = tuple(np.array(array.shape) // block_shape) + tuple(block_shape)
    new_strides = tuple(np.array(array.strides) * block_shape) + array.strides
    return as_strided(array, shape=new_shape, strides=new_strides)


def block_reduce(
    array: np.ndarray,
    block_shape: Tuple[int, ...],
    func: Callable[[np.ndarray], np.ndarray],
    **func_kwargs: Dict[str, Any],
) -> np.ndarray:
    """Reduce the array by applying a function to each block.

    Creates a view using `view_as_blocks` and applies the function to
    each block.

    Args:
        array (np.ndarray):
            The array to reduce.
        block_shape (Tuple[int, ...]):
            The shape of the blocks.
        func (Callable[[np.ndarray], np.ndarray]):
            The function to apply to each block.
        func_kwargs (Dict[str, Any]):
            Keyword arguments to pass to func.

    Returns:
        np.ndarray:
            The reduced array.
    """
    view = view_as_blocks(array, block_shape)
    return func(view, axis=tuple(range(array.ndim, view.ndim)), **func_kwargs)


def mean_pool(image: np.ndarray, pool_size: int) -> np.ndarray:
    r"""Reduce an image by applying a mean to each block.

    Uses `wsic.utils.block_reduce` to apply `np.mean` in blocks to an
    image.

    This is significantly slower than `cv2.INTER_AREA` interpolation and
    `scipy.ndimage.zoom`, but a used as fallback for when neither
    optional dependency is available.

    Note that the output shape will always round down to the nearest
    integer:

    .. math::
        \left\lfloor
        \frac{\texttt{image.shape}}{\texttt{pool_size}}
        \right\rfloor

    Args:
        image (np.ndarray):
            The image to reduce.
        pool_size (int):
            The size of the blocks to apply `np.mean` to.

    Returns:
        np.ndarray:
            The reduced image.
    """
    out_ndim = image.ndim
    image = np.atleast_3d(image)
    block_shape = (pool_size, pool_size, 1)
    reduced = block_reduce(image, block_shape, np.mean, dtype=image.dtype)
    if reduced.ndim == 3 and out_ndim == 2:
        return reduced.squeeze(axis=2)
    return reduced


def resize_array(
    array: np.ndarray,
    shape: Tuple[int, ...],
    interpolation: Union[str, int] = "bilinear",
    cv2_kwargs: Dict[str, Any] = None,
    pil_kwargs: Dict[str, Any] = None,
    zoom_kwargs: Dict[str, Any] = None,
) -> np.ndarray:
    """Resize an array (image).

    Tries to use the fastest method available by trying several
    libraries in turn. The order of preference is:
        1. `cv2.resize`
        2. 'PIL.Image.resize'
        3. `scipy.ndimage.zoom`
        4. Nearest neighbour subsampling

    Args:
        array (np.ndarray):
            The array to resize.
        shape (Tuple[int, ...]):
            The shape of the output array.
        interpolation (Union[str, int]):
            The interpolation method to use.
            Defaults to `bilinear`.
        cv2_kwargs (Dict[str, Any]):
            Keyword arguments to pass to `cv2.resize`.
        pil_kwargs (Dict[str, Any]):
            Keyword arguments to pass to `PIL.Image.resize`.
        zoom_kwargs (Dict[str, Any]):
            Keyword arguments to pass to `scipy.ndimage.zoom`.
            Defaults to `{"mode": "reflect"}`.

        Returns:
            np.ndarray:
                The resized array.
    """
    if zoom_kwargs is None:
        zoom_kwargs = {"mode": "reflect"}

    with suppress(ImportError):
        import cv2

        str_to_cv2_interpolation = {
            "nearest": cv2.INTER_NEAREST,
            "bilinear": cv2.INTER_LINEAR,
            "bicubic": cv2.INTER_CUBIC,
            "box": cv2.INTER_AREA,
            "area": cv2.INTER_AREA,
            "lanczos": cv2.INTER_LANCZOS4,
        }
        cv2_interpolation = str_to_cv2_interpolation[interpolation]

        return cv2.resize(
            array, shape[::-1], interpolation=cv2_interpolation, **(cv2_kwargs or {})
        )

    with suppress(ImportError):
        from PIL import Image

        str_to_pillow_interpolation = {
            "nearest": Image.Resampling.NEAREST,
            "bilinear": Image.Resampling.BILINEAR,
            "bicubic": Image.Resampling.BICUBIC,
            "box": Image.Resampling.BOX,
            "area": Image.Resampling.BOX,
            "lanczos": Image.Resampling.LANCZOS,
        }
        pil_interpolation = str_to_pillow_interpolation[interpolation]

        return np.array(
            Image.fromarray(array).resize(
                shape[::-1],
                resample=pil_interpolation,
                **(pil_kwargs or {}),
            )
        )

    with suppress(ImportError):
        from scipy import ndimage

        str_to_order = {
            "nearest": 0,
            "bilinear": 1,
            "bicubic": 3,
            "box": 2,
            "area": 2,
            "lanczos": 4,
        }
        order = str_to_order[interpolation]
        zoom = np.divide(shape, array.shape[:2])
        zoom = np.append(zoom, 1)

        return ndimage.zoom(
            array,
            zoom,
            order=order,
            **(zoom_kwargs or {}),
        )

    warnings.warn(
        "Neither OpenCV nor scipy are installed for image resizing. "
        "A slower and lower quality method will be used."
    )
    # Nearest neighbour sample the numpy array
    y = np.linspace(0, array.shape[0], shape[0], endpoint=False).round().astype(int)
    x = np.linspace(0, array.shape[1], shape[1], endpoint=False).round().astype(int)
    return array[np.ix_(y, x)]



"""Top-level package for wsic."""

__author__ = """John Pocock"""
__email__ = "j.c.pocock@warwick.ac.uk"
__version__ = "0.4.1"

from . import codecs, magic, metadata, multiprocessing, readers, types, writers

__all__ = [
    "codecs",
    "magic",
    "readers",
    "types",
    "writers",
    "metadata",
    "multiprocessing",
]



#!/usr/bin/env python

"""The setup script."""

from setuptools import find_packages, setup

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "numpy",
    "zarr",  # Includes numcodecs as a dependency
]

# Requirements which improve performance
performance_requirements = ["opencv-python"]

# Extra codecs support
codec_requirements = ["glymur", "imagecodecs", "qoi"]

test_requirements = [
    "pytest>=3",
    "opencv-python",
    "scipy",
    "scikit-image",
]
test_requirements += performance_requirements
test_requirements += codec_requirements

docs_requirements = [
    "sphinx",
    "sphinx-autoapi",
]

alternative_requirements = [
    "scipy",  # Alternative to scikit-image, is a dependency of scikit-image
    "scikit-image",  # Alternative to opencv-python for some operations
]

# Command-line interface requirements
cli_requirements = ["Click>=7.0"]

# User experience improving requirements
ux_requirements = ["tqdm"]

# All extra requirements
all_extra_requirements = (
    test_requirements
    + docs_requirements
    + codec_requirements
    + alternative_requirements
    + cli_requirements
    + ux_requirements
)

# Optional dependencies
extra_requirements = {
    "all": all_extra_requirements,
    "test": test_requirements,
    "docs": docs_requirements,
    "cli": cli_requirements,
    "ux": ux_requirements,
    "performance": performance_requirements,
    "codecs": codec_requirements,
    "jpeg2000": ["glymur"],
}

setup(
    author="John Pocock",
    author_email="j.c.pocock@warwick.ac.uk",
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    description="Whole Slide Image (WSI) conversion for brightfield histology images",
    entry_points={
        "console_scripts": [
            "wsic=wsic.cli:main",
        ],
    },
    install_requires=requirements,
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="wsic",
    name="wsic",
    packages=find_packages(include=["wsic", "wsic.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    extras_require=extra_requirements,
    url="https://github.com/john-p/wsic",
    version="0.4.1",
    zip_safe=False,
)
